/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/



#ifndef __DAG_H
#define __DAG_H

/*! \file DAG.h
\brief This file specifies the API of the DAG class and INFO_NODE and INFO_EDGE classes.

Directed Acyclic Graphs (DAG) are well known in graph theory and it optimizing compilation. This file contains the API of our implementation of Directed Acyclic Graphs as used in optimizing compilation. The DAG model we implement here is described in \ref dag_sec.
*/

/** \example dag_example.cpp
 * This is an example of how to use the DAG class.
 * */

/** \example RS_example.cpp
 * This is an example of how to compute the register saturation of a DAG.
 * */

 
 
/** \internal
\def MAX_NODE 
The maximal number of nodes inside a DDG is statically limited to 10000.
*/
#ifndef MAX_NODE
#define MAX_NODE 10001
#endif 




#include <string> 
#include <iostream>
#include <fstream>
#include "LEDA/graph/graph.h" /* GRAPH, string */
#include "LEDA/core/set.h"
#include "LEDA/graph/node_set.h"
#include "LEDA/graph/node_array.h"
#include "LEDA/graph/node_matrix.h" /* GRAPH, string */
#include "LEDA/core/array.h"
#include "LEDA/graph/node_list.h"
#include "LEDA/core/d_array.h"
#include "LEDA/graph/mcb_matching.h" /* bipartie matching algorithm */
#include "LEDA/graph/basic_graph_alg.h" /* transitive closure */
#include <LEDA/core/map2.h>
#include "ddg_exceptions.h"
#include "architecture.h"
#include "info_edge.h"
#include "info_node.h"

#ifndef MAX_STRING
#define MAX_STRING 1000
#endif

/*! \namespace DDG
\brief All the API symbols used in the DDG library.
*/





using namespace std;
using namespace LEDA;
namespace DDG{


/** \internal
 * \fn template<class T> LEDA::set<T> list_to_set(const LEDA::list<T> l);
* \internal
\brief Converts a LEDA list to a LEDA set.
\param[in] l  the LEDA list to convert
 * \returns a LEDA set
*/
			 
template<class T> LEDA::set<T> list_to_set(const LEDA::list<T> l);

/** \internal
 * \fn template<class T> LEDA::list<T> set_to_list(const set<T> l);
\brief Converts a LEDA set to a LEDA list.
\param[in] l the LEDA set to convert
 * \returns a LEDA linear list
*/
template<class T> LEDA::list<T> set_to_list(const set<T> l) ;

template<class T> set<T> list_to_set(const list<T> l) {
	T m;
	set<T> tmp;
	forall(m , l){tmp.insert(m);}
	return tmp;
}

template<class T> list<T> set_to_list(const set<T> l) {
	T m;
	list<T> tmp;
	forall(m , l){tmp.append(m);}
	return tmp;
}


/*! \fn bool lt(node u, node v);
 * @param[in] u A first node
 * @param[in] v A second node
 * @return Returns True if \f$ u < v \f$, i.e., \f$ u \neq \wedge \exists \textrm{ a path } (u,\ldots,v) \f$.
  \pre  \f$ u \f$ and \f$ v \f$ belongs to the same DAG.
  This is a notation from the lattices and order theory
 */
bool lt(node u, node v);


/*! \fn bool le(node u, node v);
 * @param[in] u A first node
 * @param[in] v A second node
 * @return Returns True if \f$ u \le v \f$, i.e., \f$  u=v \vee \exists \textrm{ a path } (u,\ldots,v) \f$.
  \pre  \f$ u \f$ and \f$ v \f$ belongs to the same DAG.
  This is a notation from the lattices and order theory
    \remarks  this relationship is reflexive: \f$ \forall u \in V, u \le u \f$

 */
bool le(node u, node v);


/*! \fn bool gt(node u, node v);
 * @param[in] u A first node
 * @param[in] v A second node
 * @return Returns True if \f$ u > v \f$, i.e., \f$ u \neq \wedge \exists \textrm{ a path } (v,\ldots,u) \f$.
  \pre  \f$ u \f$ and \f$ v \f$ belongs to the same DAG.
  This is a notation from the lattices and order theory
 */


bool gt(node u, node v);


/*! \fn bool ge(node u, node v);
 * @param[in] u A first node
 * @param[in] v A second node
 * @return Returns True if \f$ u \ge v \f$, i.e., \f$ u=v \vee \exists \textrm{ a path } (v,\ldots,u) \f$.
  \pre  \f$ u \f$ and \f$ v \f$ belongs to the same DAG.
  This is a notation from the lattices and order theory
    \remarks  this relationship is reflexive: \f$ \forall u \in V, u \ge u \f$
 */

bool ge(node u, node v);

/*! \fn bool parallel(node u, node v);
 * @param[in] u A first node
 * @param[in] v A second node
 * @return True if  \f$ u || v \f$, i.e., \f$  \neg(u \le v)  \wedge \neg (v \le u) \f$.
 \pre  \f$ u \f$ and \f$ v \f$ belongs to the same DAG.
* This is a notation from the lattices and order theory
 */
bool parallel(node u, node v);

/*! \fn bool comparable(node u, node v);
 * @param[in] u A first node
 * @param[in] v A second node
 * @return True if  \f$ u \sim v \f$, i.e., \f$  u\le v  \vee v \le u \f$.
 \pre  \f$ u \f$ and \f$ v \f$ belongs to the same DAG.
* This is a notation from the lattices and order theory
  \remarks  \f$ \forall u \in V, u \sim u \f$

 */
bool comparable(node u, node v);

/*! \fn int node_compare(const node &u, const node &v)
 * This function is used for sorting.
 * @param[in] u A first node
 * @param[in] v A second node
 * @return -1 if \f$ u < v \f$. 0 if \f$ u = v\f$. +1 if \f$ u > v\f$
 \pre  \f$ u \f$ and \f$ v \f$ belongs to the same DAG object, and they are comparable: \f$ u \sim v \f$. The internal state of the DAG object should be consistent, see DAG::build_internal_structures(). 
 */
int compare_nodes(const node &u, const node &v);



class INFO_NODE;
class INFO_EDGE;
class DAG;
class LOOP;






/*! \class DAG
	\brief This class represents a directed acyclic graph (DAG). 
	\author Sid Touati		
 * 
	It inherits from the LEDA class GRAPH<INFO_NODE, INFO_EDGE>. The attributes of the nodes are of type INFO_NODE, and the attributes of edges are of type INFO_EDGE. The implemented DAG model is described in \ref dag_sec.
	
	For more information about LEDA, please consult http://www.algorithmic-solutions.info/leda_manual


*/
class DAG: public GRAPH <INFO_NODE, INFO_EDGE>
{
	protected:
		unsigned int cpt_nodes; // a counter incremented each new node (but never decremented). Used for assigning a unique identifier per node.
	
	
	/*useful sets*/
	LEDA::d_array<int, LEDA::set<node> > V_R; /* set V_R,t: set of values in G for each register type */
	LEDA::d_array<int, LEDA::set<edge> > E_R; /* The set of flow edges through registers for each register type */
	
	ARCHITECTURE ISA;
	
	
	node_array<set<node> > down;
	node_array<set<node> > up;
	node_array<int> longuest_path_from;
	node_array<int> longuest_path_to;
	node_matrix<int> lp_array; /* lp_array(u,v)=lp(u,v) */
	/*shortest path to and from a node */
	node_array<int> shortest_path_from;
	node_array<int> shortest_path_to;
	/* sources and targets of a node */
	set<node> Sources; /* sources of the DAG */
	set<node> Targets; /* leaf of the DAG */
	map2<node, int, set<node> > Cons; /* consumer set of a value of a given register type Cons[u,t] */
	LEDA::map<int,node> node_index_table; //save the id of each node : node_index_table[u] is the id of the node u
 public:
	 
	 /**  \name Creation */
   	//@{
	 /** \brief  Declares an empry DAG.
	  * 
	  * Default constructor.
	  * 
	  * It creates an empty DAG. A default architecture (ISA) is assumed, but the user may attach another ISA to the DAG. See DDG::DAG::set_ISA */
	 DAG(); 
	 
	/**Constructs a DAG after reading the generic description of the architecture (ISA). 
	  * \param[in] ARCHITECTURE The ARCHITECTURE (ISA) object containing the ISA description. 
   	
	  */
  	DAG(const ARCHITECTURE); 
  	
  	~DAG();
	//@}
	
	/**  \name Access Operations */
   	//@{
	/** \brief Returns the set of all edges from the source node \f$ u \f$ to the destination node \f$ v \f$.  */
  	LEDA::set<edge> edges_from_to(node u, node v) const;
	
   	 /** \brief Returns the set of all edges from the source node  to the destination node.
	* \param[in] id_u The integer identifier of the source node.
   	* \param[in] id_v The integer identifier of the destination node.
      * */
  	LEDA::set<edge> edges_from_to(int id_u, int id_v) const;

	/** \brief Returns the value of the longest path of the DAG. */
  	int LongestPath() const ;
 
     	/** \brief Returns the value of the longest path of the DAG. 
	*  \param[out] nl Contains the list of nodes belonging to the computed longest path. 
	* \warning Despite the unicity of its value, the longest path in a DAG may not be unique. 
        */
  	int LongestPath(LEDA::list<node> &nl) const;
	
	/** \brief Returns the value of the longest path of the DAG.
	*  \param[out] el Contains the list of edges belonging to the computed longest path.
	* \warning Despite the unicity of its value, the longest path in a DAG may not be unique.
        */
	int LongestPath(list<edge> &el) const;
  
		   /** \brief This is an alias of the method DAG::LongestPath(). It returns the value of the longest (critical) path */
  	int CriticalPath();
    
 	/** \brief Returns the value of the longest (critical) path. 
   	 *  \param[out] nl Contains the list of nodes belonging to the computed longest path. 
  	* \warning The longest path in a DAG may not be unique
   	* */
  	int CriticalPath(LEDA::list<node> &nl);
					      

	/** \brief Returns the value of the longest (critical) path.
   	 *  \param[out] el Contains the list of edges belonging to the computed longest path.
  	* \warning The longest path in a DAG may not be unique
   	* */
	int CriticalPath(LEDA::list<edge> &el);
	
	/** \brief Returns the value of the longest path from the top of the DAG (sources) up to a node given as argument */
  	int LongestPathTo(node) const;
	
	/** \brief Returns the value of the longest path from a node given as  argument to the bottom the DAG (sinks) */
	int LongestPathFrom(node) const ;
	
	/** \brief Returns the value of the shortest path from the top of the DAG (sources) up to a node given as argument */
  	int ShortestPathTo(node) const;
	
	/** \brief Returns the value of the shortest path from a node given as  argument to the bottom the DAG (sinks) */
  	int ShortestPathFrom(node) const;
	
	/** \brief Returns the value of the longest path from a node \f$ u \f$ to a node \f$ v \f$ given as  arguments*
	\pre Such path exists
	*/
  	int lp(node u, node v) const ;
  	
	/** \brief Returns the set \f$ \downarrow u\f$ (the set of all descending nodes)., see  \ref dag_notations. */
  	LEDA::set<node> get_down(node u) const;
	
	/** \brief Returns the set \f$ \uparrow u \f$.(the set of all ascenfing nodes), see  \ref dag_notations */
  	LEDA::set<node> get_up(node u) const;
  
	
	/** \brief Returns the set \f$ \Gamma^+_G(u) \f$ (set of children), see  \ref dag_notations. */
  	LEDA::set<node> get_parents(node u) const;
	
	/** \brief Returns the set \f$ \Gamma^-_G(u) \f$ (set of parents), see  \ref dag_notations. 
	*/
  	LEDA::set<node> get_children(const node u) const;
		   
 	 /** \brief Returns the node with the integer identifier given as argument. 
   	* \throws IllegalAccessNode If the node id does not exist, or if the node is hidden, then an exception of type DDG:IllegalAccessNode is thrown.
	   * */
  	node node_with_id(int i) const;

    /** \brief Returns the node with the name given as argument. 
   	* \throws IllegalAccessNode If the node id does not exist, or if the node is hidden.
	   * */
    node node_with_name(std::string name) const;
    
	/** \brief Returns true if the node is an instruction writing into a register (i.e. if it belongs to the \f$ V_R \f$ set).
	\pre There is a unique register type in the architecture
	\throws NonUniqueRegisterType
	*/
  	bool is_value(node n) const; 
		   
	/** \brief Returns true if the node defined by its indentifier is an instruction writing into a register of type t (i.e. if it belongs to the \f$ {V_R,t} \f$ set) 
	*/
  	bool is_value(int i, int t)const ;
		   
	/** \brief Returns true if the node is an instruction writing into a register of type t (i.e. if it belongs to the \f$ V_{R,t} \f$ set).
	*/
  	bool is_value(node n, int t) const; 
	   
	/** \brief Returns true if the node defined by its indentifier is an instruction writing into a register (i.e. if it belongs to the \f$ V_R \f$ set) 
	\pre There is a unique register type in the architecture
	\throws NonUniqueRegisterType
	*/
  	bool is_value(int i)const ;
  	
	/** \brief Returns true if the edges is a flow dependence (not necessarily through registers).*/
	bool is_flow(edge e) const {INFO_EDGE ie=inf(e); return ie.is_flow(); }

	/** \brief Returns true if the edges is a flow dependence through a register. This is the case of a flow edge having a source writing into a register.
	\pre There is a unique register type in the architecture
	\throws NonUniqueRegisterType
	*/
	bool is_flow_reg(edge e) const {
	INFO_EDGE ie=inf(e);	                     
	return ie.is_flow_reg(ISA.get_the_unique_register_type_id());
	}
	
	/** \brief Returns the register type of a flow edge
	\pre The edge should be a flow through register (flowdep_reg)
	 */
    
	int get_register_type_id(edge e) const;
	
	/** \brief Returns true if the edges is a flow dependence through a register of type t. This is the case of a flow edge having a source writing into a register.*/
	bool is_flow_reg(edge e, int t) const {
	INFO_EDGE ie=inf(e);	                     
	return ie.is_flow_reg(t);
	}
		   
	/** \brief Returns the type of the edge \f$ e \f$. (flow, serial, antidep, ...). */
	DDG::edge_type etype(edge e)const ;

		   
  	/** \brief Returns the set \f$ V_R \subseteq V \f$ of instructions (nodes) writing into registers. 
  	\pre a unique register type should exist in the architecture
  	\throw NonUniqueRegisterType
  	*/
  	set<node> get_V_R() const
  	{return V_R[ISA.get_the_unique_register_type_id()];}
  	
  	/** \brief Returns the set \f$ V_ {R,t}\subseteq V \f$ of instructions (nodes) writing into registers of type t (given by its id). 
  	\throw BadRegType
  	*/
  	set<node> get_V_R(int t) const
  	{
  		if (ISA.exist_register_type_id(t))
  			return V_R[t];
  		else throw BadRegType();
  	}
  	
	LEDA::list<REGISTER_TYPES> T() const ;
  	/** \brief Returns the set \f$ E_R \subseteq E \f$ of flow dependence edges through registers. 
  	 \pre A unique register type must exist. 
  	 \throw NonUniqueRegisterType
  	 */
  	LEDA::set<edge> get_E_R() const
  	{return E_R[ISA.get_the_unique_register_type_id()];}
  	
  	/** \brief Returns the set \f$ E_{R,t} \subseteq E \f$ of flow dependence edges through registers of type t (given by its id). 
  	\throw BadRegType
  	 */
  	LEDA::set<edge> get_E_R(int t) const
  	{
  		if (ISA.exist_register_type_id(t))
  			return E_R[t];
  		else throw BadRegType();
  	}
    
	/** \brief Returns the set of nodes which are the sources of the DAG.  */
  	LEDA::set<node>  get_Sources() const;
	
	/** \brief Returns the set of nodes which are the targetsof (sinks) of the DAG. */
  	LEDA::set<node>  get_Targets()const ;

	/** \brief Returns the set of consumers of a node. 
	   \pre A unique register type should exist in the architecture. 
	   \throws NonUniqueRegisterType
	
	*/
	LEDA::set<node>  get_Cons(node u) const{
		return Cons(u, ISA.get_the_unique_register_type_id());
	}
	
	/** \brief Returns the set of consumers of a node writing inside a register of a given type. 	
	\pre id_regtype is an id of register type in the architetcure
	\throw BadRegType
	*/
	LEDA::set<node>  get_Cons(node u, int id_regtype) const{
		if (ISA.exist_register_type_id(id_regtype))
			return Cons(u, id_regtype);
		else throw BadRegType();
	}
	
	
	/** \brief Returns the set of nodes identifiers which are the sources of the DAG.  */
  	LEDA::set<int >  get_Sources_id() const;
	
	/** \brief Returns the set of nodes identifiers which are the targetsof (sinks) the DAG.  */
 	 LEDA::set<int >  get_Targets_id() const;
  	
	 /**  \brief Returns the set of nodes identifiers. */
	 LEDA::set<int> all_nodes_id();
		     

	/**
	 * \brief Returns the integer identifier of the edge source.    
	 * */
    	int source_id(edge e) const
    	{
	    return id(source(e));
    	}

   	

    	/** \brief Returns the integer identifier of the edge target.   
     	*  */
    	int target_id(edge e) const
    	{
	    return id(target(e));
    	}
	
	
	/** \brief Returns the ARCHITECTURE of the DAG.*/
	ARCHITECTURE get_ISA() const
	{
		return ISA;
	}


	/**
	 * \brief Returns the latency \f$ \delta(e) \f$ of the edge \f$ e \f$ 
	 * */
	int delta(edge e) const;


	/**
	 * \brief Returns the latency \f$ latency(n) \f$ of the generic instruction of the node \f$ n \f$
	 * */
	int latency (node n) const ;

	/**
	 * \brief Returns the reading latency \f$ \delta_r(n) \f$ of  the node \f$ n \f$
	 * */
	int delta_r (node n) const;
	
	/**
	 * \brief Returns the reading latency \f$ \delta_r(idn) \f$ of the node given by its id
	 * */
	int delta_r (int idn) const;
	
	/**
	 * \brief Returns the writing latency \f$ \delta_w(n) \f$ of the node \f$ n \f$
	 \pre a unique register type should exist in the architecture
	 * */
	int delta_w (node n) const;

	/**
	 * \brief Returns the writing latency \f$ \delta_w(idn) \f$ of the node given by its id
	 \pre a unique register type should exist in the architecture
	 * */
	int delta_w (int idn) const;
	
	/**
	 * \brief Returns the writing latency \f$ \delta_{w,t}(n) \f$ of the node \f$ n \f$
	 * */
	int delta_w (node n, int regtype_id) const;

	/**
	 * \brief Returns the writing latency \f$ \delta_{w,t}(idn) \f$ of the node given by its id
	 * */
	int delta_w (int idn, int regtype_id) const;
     
	/** \brief Returns the integer  identifier of the node \f$ n \f$.
	 * */
    	int id(node n) const
    	{
	    INFO_NODE in=inf(n);
	    return in.id();
    	}

    	/** \brief Returns an INSTRUCTIONS_TYPES object attached to the node \f$ n \f$. 
     
     * */
	INSTRUCTIONS_TYPES instruction_type(node n)  const;


    	/** \brief Returns an INSTRUCTIONS_TYPES object attached to the node given by its \f$ id \f$.
     
     * */
	INSTRUCTIONS_TYPES instruction_type(int id)  const;
	
	/** \brief Returns the textual attribute of the node \f$ n \f$.  */
	std::string get_string_attribute(node n) const;

	/* ------------------ Some LEDA methos ------------------- */
	/** \brief  Returns the information of node \f$ v \f$.*/
	const INFO_NODE& inf(node v) const { return LEDA_CONST_ACCESS(INFO_NODE,graph::inf(v));}

	/** \brief Returns a reference to the information of \f$ v \f$.*/
	const INFO_NODE& operator[] (node v) const
	{ return inf(v);     }

	INFO_NODE& operator[] (node v)
	{ return LEDA_ACCESS(INFO_NODE,entry(v));}
	
	/** \brief  Returns the information of node \f$ e \f$.*/
	const INFO_EDGE& inf(edge e) const
	{ return LEDA_CONST_ACCESS(INFO_EDGE,graph::inf(e));}

	/** \brief Returns a reference to the information of \f$ e \f$.*/
	const INFO_EDGE& operator[] (edge e) const
	{ return inf(e);                      }
	INFO_EDGE&       operator[] (edge e)
	{ return LEDA_ACCESS(INFO_EDGE, entry(e)); }

	/** \brief Returns the number of edges leaving to node \f$ v  \f$. */
	int  outdeg(node v)    const
	{return GRAPH<INFO_NODE,INFO_EDGE>::outdeg(v);}

	/** \brief Returns the number of edges entering \f$ v \f$. */
   	int  indeg(node v)     const
	{return GRAPH<INFO_NODE,INFO_EDGE>::indeg(v);}

	/** \brief Returns  \f$ outdeg(v) + indeg(v) \f$. */
   	int  degree(node v)    const {return (
   	GRAPH<INFO_NODE,INFO_EDGE>::indeg(v) +
   	GRAPH<INFO_NODE,INFO_EDGE>::outdeg(v));}

	/** \brief Returns the source node of edge \f$ e \f$. */
   	node source(edge e)    const
	{return GRAPH<INFO_NODE,INFO_EDGE>::source(e);}

	/** \brief Returns the target node of edge \f$ e \f$.*/
   	node target(edge e)    const
	{return GRAPH<INFO_NODE,INFO_EDGE>::target(e);}

	/** \brief Returns \f$ target(e) \f$ if \f$ v = source(e) \f$ and \f$ source(e) \f$ otherwise. */
   	node opposite(node v, edge e) const
	{return GRAPH<INFO_NODE,INFO_EDGE>::opposite(v,e);}

	/** \brief Returns \f$ target(e) \f$ if \f$ v = source(e) \f$ and \f$ source(e) \f$ otherwise. */
   	node opposite(edge e, node v) const
	{return GRAPH<INFO_NODE,INFO_EDGE>::opposite(e,v);}

	/** \brief Returns the number of nodes in the DAG */
   	int  number_of_nodes() const
	{return GRAPH<INFO_NODE,INFO_EDGE>::number_of_nodes();}

	/** \brief Returns the number of edges in the DAG. */
   	int  number_of_edges() const
	{return GRAPH<INFO_NODE,INFO_EDGE>::number_of_edges();}

	/** \brief Returns the list of all nodes in the DAG. */
	const LEDA::list<node>& all_nodes()  const
	{return GRAPH<INFO_NODE,INFO_EDGE>::all_nodes();}

	/** \brief Returns the list of all edges in the DAG. */
	const LEDA::list<edge>& all_edges()  const
	{return GRAPH<INFO_NODE,INFO_EDGE>::all_edges();}


	/** \brief Returns the list of edges leaving the node \f$ v \f$. */
   	LEDA::list<edge> out_edges(node v) const
	{return GRAPH<INFO_NODE,INFO_EDGE>::out_edges(v);}

	/** \brief Returns the list of edges entering the node \f$ v \f$. */
   	LEDA::list<edge> in_edges(node v) const
	{return GRAPH<INFO_NODE,INFO_EDGE>::in_edges(v);}


	/** \brief Returns the first node in the DAG. */
   	node first_node()      const
	{return GRAPH<INFO_NODE,INFO_EDGE>::first_node();}

	/** \brief Returns the last node in the DAG. */
   	node last_node()      const
	{return GRAPH<INFO_NODE,INFO_EDGE>::last_node();}

	/** \brief Returns a random node of the DAG (nil if empty). */
   	node choose_node()     const
	{return GRAPH<INFO_NODE,INFO_EDGE>::choose_node();}

	/** \brief Returns the successor of node \f$ v \f$ (nil if it does not exist). */
   	node succ_node(node v) const
	{return GRAPH<INFO_NODE,INFO_EDGE>::succ_node(v);}
   	
	/** \brief Returns the predecessor of node \f$ v \f$ (nil if it does not exist). */
   	node pred_node(node v) const
	{return GRAPH<INFO_NODE,INFO_EDGE>::pred_node(v);}

	/** \brief Returns the first edge in the DAG. */
   	edge first_edge()      const
	{return GRAPH<INFO_NODE,INFO_EDGE>::first_edge();}

	/** \brief Returns the last edge in the DAG. */
   	edge last_edge()      const
	{return GRAPH<INFO_NODE,INFO_EDGE>::last_edge();}

	/** \brief Returns a random edge of the DAG (nil if empty). */
   	edge choose_edge()     const
	{return GRAPH<INFO_NODE,INFO_EDGE>::choose_edge();}

	/** \brief Returns the successor of edge \f$ e \f$ (nil if it does not exist). */
	edge succ_edge(edge e) const
	{return GRAPH<INFO_NODE,INFO_EDGE>::succ_edge(e);}

	/** \brief Returns the predecessor of edge \f$ e \f$ (nil if it does not exist). */
	edge pred_edge(edge e) const
	{return GRAPH<INFO_NODE,INFO_EDGE>::pred_edge(e);}

	/** \brief Returns true if the DAG is empty */
	bool empty() const
	{ return GRAPH<INFO_NODE,INFO_EDGE>::empty(); }

	/** \brief Returns true if the node \f$ v \f$ belongs to the DAG. */ 
	bool   member(node v) const
	{ return GRAPH<INFO_NODE,INFO_EDGE>::member(v); }

	/** \brief Returns true if the edge \f$ e \f$ belongs to the DAG. */
	bool   member(edge e) const
	{ return  GRAPH<INFO_NODE,INFO_EDGE>::member(e); }
	
	/** \brief  Returns true if \f$ e \f$ is hidden, false otherwise.
	See DAG::hide_edge(edge e).
	*/
   	bool is_hidden(edge e) const
	{ return GRAPH<INFO_NODE, INFO_EDGE>::is_hidden(e); }

   	/** \brief  Returns true if \f$ v \f$ is hidden, false otherwise.
	See DAG::hide_node(node v).
	*/
   	bool is_hidden(node v) const
	{ return GRAPH<INFO_NODE, INFO_EDGE>::is_hidden(v); }

   	/** \brief Returns the list of all hidden edges. */
   	LEDA::list<edge> hidden_edges() const
	{return GRAPH<INFO_NODE, INFO_EDGE>::hidden_edges();}

   	/** \brief Returns the list of all hidden nodes. */
   	LEDA::list<node> hidden_nodes() const
	{return GRAPH<INFO_NODE, INFO_EDGE>::hidden_nodes();}

	/** \brief Returns the maximal edge latency between to nodes, i.e., \f$ \max_{e=(u,v)} \delta(e) \f$. */
	int max_delta(node u, node v) const;

	/** \brief Returns the maximal edge latency between to nodes (given by their ids), i.e., \f$ \max_{e=(u,v)} \delta(e) \f$. */
	int max_delta(int u, int v) const
	{return max_delta(node_with_id(u),node_with_id(v));}

	/** \brief Returns the minimal edge latency between to nodes, i.e., \f$ \min_{e=(u,v)} \delta(e) \f$. */
	int min_delta(node u, node v) const;
	
	/** \brief Returns the minimal edge latency between to nodes (given by their ids), i.e., \f$ \min_{e=(u,v)} \delta(e) \f$. */
	int min_delta(int u, int v) const
	{return min_delta(node_with_id(u),node_with_id(v));}

	/** \brief Looks for the instruction type (inside the DAG ISA) that has the opcode given as argument. If not found, it returns an INSTRUCTIONS_TYPES object with opcode_id equal to -1. */
	INSTRUCTIONS_TYPES search_instruction_type(const std::string  opcode) const;

	/** \brief Looks for the instruction type (inside the DAG ISA) that has the opcode id given as argument. Returns NULL if not found. If not found, it returns an INSTRUCTIONS_TYPES object with opcode_id equal to -1. */
	INSTRUCTIONS_TYPES search_instruction_type(int opcode_id) const;
	
	/** \brief A simple check method that returns true if the DAG object seems ok. False otherwise. */
	bool check();
	/** \brief A simple check method that returns true if the DAG object seems ok for the specified register type. False otherwise. */
	bool check(int regtype_id);
	
	/** \brief Returns the node flag of n. */
	NodeFlag node_flag ( node n) const;
	
	/** \brief Returns the list of register types of the ISA of the DAG. The ISA object should be attached to the DAG before calling this function. See the DAG construtor method, or the DDG:DAG:set_ISA method.
	LEDA::list<REGISTER_TYPES> T() const ;
	//@}
	 
	/** \name Update Operations */
	//@{			  
				  
  	/** \brief Empty the DAG. */
  	void clear();
		  
  	/** \brief Creates a new node.
	* \param[in] it_id The id of the INSTRUCTIONS_TYPES object describing the generic instruction of the node
	 * \return The integer node identifier.
	 * \remarks This method assigns a unique inter identifier for the created node. 
	 * 
	 * See also DAG::build_internal_structures().
   	*/
  	int new_node(int it_id);

	/** \brief Creates a new node.
	* \param[in] it_id The id of the INSTRUCTIONS_TYPES object describing the generic instruction of the node
	 * \param[in] text_inst A textual attribute of the instruction
	 * \return The integer node identifier.
	 * \remarks This method assigns a unique inter identifier for the created node.
	 *
	 * See also DAG::build_internal_structures().
   	*/
   	  	
	int new_node(int it_id, std::string text_inst);
	
	
	/** \brief Creates a new node.
	* \param[in] it_id The id of the INSTRUCTIONS_TYPES object describing the generic instruction of the node
	  \param[in] n_id A unique id for the node to create
	 * \return The integer node identifier. Returns -1 if the there exists an node with the input id.
	 \throws NonUniqueNodeID if the input node id already exists
	 * \remarks This method creates a node with an identifier given as input. 
	 \pre n_id is positive
	
	 * 
	 * See also DAG::build_internal_structures().
   	*/
  	int new_node(int it_id, int n_id);
  	
  	
  	/** \brief Creates a new node.
	* \param[in] it_id The id of the INSTRUCTIONS_TYPES object describing the generic instruction of the node
	  \param[in] n_id A unique id for the node to create
	  \param[in] text_inst A textual attribute of the instruction
	 * \return The integer node identifier. 
	 \throws NonUniqueNodeID if the input node id already exists
	 * \remarks This method creates a node with an identifier given as input. 
	 \pre n_id is positive
	
	 * 
	 * See also DAG::build_internal_structures().
   	*/
  	int new_node(int it_id, int n_id, std::string text_inst);
					  
	/** \brief Deletes the node given as argument.  
	 * 
	 * All the edges incident to the deleted node are removed from the DAG too.
	*/
  	void del_node(node); 
	
	/** \brief Deletes the node whose id is given as argument. 
	 * 
	 * All the edges incident to the deleted node are removed from the DAG too.
	 * 
	* See also DAG::build_internal_structures().
	*/
  	void del_node(int node_id);
  
  	/** \brief Creates a new edge. 
	 * 
	* It takes three arguments. If flow dependence, the inserted edge is added to the set \f$ E_R \f$.
	* \param[in] s The integer identifier of the source node.
   	* \param [in] t The integer identifier of the  destination node.
   	* \param[in]  ie The INFO_EDGE object describing the edge attribute.
	 * See also DAG::build_internal_structures().
	 */
	edge new_edge(int s, int t, INFO_EDGE ie);
	
  	/** \brief Creates a new edge. 
	 * 
	* It takes three arguments. If flow dependence, the inserted edge is added to the set \f$ E_R \f$.
   	* \param[in] s The  source node.
   	* \param [in] t The destination node.
   	* \param[in]  ie The INFO_EDGE object describing the edge attribute.
	* See also DAG::build_internal_structures().
	* */
  	edge new_edge(node s, node t, INFO_EDGE ie); 
  
	/** \brief Deletes the edge given as argument. */
  	void del_edge(edge e); 

  	
  	/** \brief Duplicates a DAG. 
	\param[in] G2 The current DAG object will contain a duplicated copy of  G2. This method will keep the same node ids, same edge and node information, etc. The internal consistency of the duplicated DAG is guaranteed.
	 \remarks
	 DAG::build_internal_structures() is called inside this method.
	  */
  	void copy(const DAG& G2);
	
	
	/** \brief Duplicates a DAG. 
		\param[in] G2 The current DAG object will contain a duplicated copy of  G2. This method will keep the same node ids, same edge and node information, etc. The internal consistency of the duplicated DAG is guaranteed.
	 * \param[out] mn mn is a mapping from the set of nodes of G2 to the set of nodes of the current DAG. \f$ \forall u\in G_2, mn[u] \f$ contains the node belonging to the current DAG corresponding to its original copy \f$ u \in G_2 \f$.
	 * \param[out] me me is a mapping from the set of edges of G2 to the set of edges of the current DAG. \f$ \forall e \in G_2, me[e] \f$ contains the edge belonging to the current DAG corresponding to its original copy \f$ e \in G_2 \f$.
	 \remarks
	   DAG::build_internal_structures() is called inside this method.
	 */
	void copy(const DAG& G2, node_map<node> &mn, edge_map<edge> &me);

	/** \brief Retrieves a loop body. 
		\param[in] G2 The current DAG object will contain the loop body of G2. This method will keep the same node ids, same edge and node information, etc. The internal consistency of the DAG is guaranteed.
	 * \param[out] mn mn is a mapping from the set of nodes of G2 to the set of nodes of the current DAG. \f$ \forall u\in G_2, mn[u] \f$ contains the node belonging to the current DAG corresponding to its original copy \f$ u \in G_2 \f$.
	 * \param[out] me me is a mapping from the set of edges of G2 to the set of edges of the current DAG. \f$ \forall e \in G_2, me[e] \f$ contains the edge belonging to the current DAG corresponding to its original copy \f$ e \in G_2 \f$.
	 \remarks
	   DAG::build_internal_structures() is called inside this method.
	   The loop body contains the set of all nodes, but the set of edges contains only the edges with null distance: \f$ \forall e \textrm{ in the loop body}, \lambda(e)=0 \f$.
	 */
  	void get_loop_body(const LOOP& G2, node_map<node> &mn, edge_map<edge> &me);

  	/** \brief Retrieves a loop body. 
	\param[in] G2 The current DAG object will contain the loop body of  G2. This method will keep the same node ids, same edge and node information, etc. The internal consistency of the duplicated DAG is guaranteed.
	 \remarks
	 DAG::build_internal_structures() is called inside this method.
	 The loop body contains the set of all nodes, but the set of edges contains only the edges with null distance: \f$ \forall e \textrm{ in the loop body}, \lambda(e)=0 \f$.
	  */
  	void get_loop_body(const LOOP& G2);
  	
	/** \brief Build internal data structures for the DAG. 
	 * 
	* For internal information consistency, this function member should be called after update operations such as DAG::new_node(INSTRUCTIONS_TYPES it), DAG::new_edge(INFO_EDGE), etc. This method compute some internal information returned by other methods (the set \f$ V_R \f$, the set of consumers per value, all pair shortest paths, transitive closure, set of descendant and ascendant nodes, set of values, etc.). If the DAG is not modified after calling build_internal_structures(), all its internal information remain consistent.
	\remarks For efficiency, the user can group series of update operations before ultimately calling build_internal_structures() 
  	\pre The graph should not contain a circuit. 
	*/
  	void build_internal_structures(); 
 	
	/** \brief  Sets the set of generic instruction types (ISA) encapsulated inside the DAG 
	 * 
	*  \param[in] isa An ARCHITECTURE object describing a user-defined ISA.
	* \warning It is better to clear the DAG before changing its ISA. If the ISA of a non empty DAG is modified on-the-fly, the successive behavior of the DAG object would be indefined.	
	 */
	void set_ISA( ARCHITECTURE isa )   
	{
		ISA = isa;
	}

    	/** \brief  Sets a latency \f$ \delta(e) \f$ to the edge \f$ e \f$.  
    	 * */
	void set_delta(edge e, int l);
    
    	/** \brief Sets an INSTRUCTIONS_TYPES object (given by its opcode id) attached to the node \f$ n \f$.
     	* */
	void  set_instruction_type(node n, int it_id);
	
	/** \brief Sets a textual attribute to the node  \f$ n \f$.
	 * For instance, this textual attribute can be used to associate the textual code of the instruction.
	 * \remarks
	 * 	- Currently, a textual attribute is a single textual line. Hence, it should not contain the newline character \\n. If they exist, the current implementation replaces them by white spaces.
	 * 	- Double quotes characters " should be specialized. They should be written as \" inside the string attribute. */
	
	void set_string_attribute(node u, std::string inst_text);
	
	

	/** \brief Sets an edge type (flow, serial, antidep, etc.).
     	* */
	void  set_etype(edge e, DDG::edge_type et);


	/*------------------- Some LEDA Methods --------------*/
	/** \brief Removes temporarily the edge \f$ e \f$.
	The edge can be restored by DAG::restore_edge(edge e).
	 * \remarks A hidden edge does no longer  belong to the list of graph edges
	*/
	void hide_edge(edge e);
	
	/** \brief Hides all edges in the list el.
	 * \remarks A hidden edge does no longer  belong to the list of graph edges
	 */
	void hide_edges(const LEDA::list<edge>& el) ;

	/** \brief Restores the hidden edge \f$ e \f$ .
	 * \warning If an endpoint of \f$ e \f$ is hidden, an exception occurs.
	 */
	void restore_edge(edge e); 

	/** \brief Restores all edges in the list el.
	 * \warning If an endpoint of an edge is hidden, an exception occurs.
	 */
	void restore_edges(const LEDA::list<edge>& el) ;
	
	/** \brief Restores all hidden edges.
	 * \warning If an endpoint of an edge is hidden, an exception occurs.
	 */
	void restore_all_edges(); 

	/** \brief Removes temporarily the node \f$ v \f$.
	Leaving and entering edges are also hidden.
	The node can be restored by DAG::restore_node(node v).
	 * \remarks A hidden node does no longer belong to the list of graph nodes
	*/
	void hide_node(node v);

	/** \brief Hides all nodes in the list nl.
	 * \remarks  A hidden node does no longer belong to the list of graph nodes
	 */
	void hide_nodes(const LEDA::list<node>& nl); 

	/** \brief Removes temporarily the node \f$ v \f$.
	Leaving and entering edges are also hidden. The node can be restored by DAG::restore_node(node n).

	\param[in] v the node to hide.
	\param[out] h_edges the list of leaving and entering edges hidden in consequence of hiding the node \f$ v \f$.
	 * \remarks  A hidden node does no longer belong to the list of graph nodes
	*/
	void hide_node(node v, LEDA::list<edge>& h_edges) ;
	


	/** \brief Restores the hidden node \f$ v \f$. Note that no edge adjacent to v that was hidden by G.hide_node(v) is restored by this operation.*/
	void restore_node(node v) ; 

	/** \brief Restores all hidden nodes. Note that no adjacent edges to the hidden nodes are restored by this operation. */
	void restore_all_nodes(); 


	/** \brief Deletes all edges from the DAG.

	See also DAG::empty();

	*/
	void del_all_edges();


	/** \brief  Moves the edge \f$ e \f$ to source \f$ v \f$ and target \f$ w \f$  */
   	void move_edge(edge e, node v, node w){GRAPH<INFO_NODE, INFO_EDGE>::move_edge(e,v,w);}

	/** \brief Set the flag of the node n. */
	void set_node_flag ( node n, const NodeFlag theValue );
	//@}
	
	
	/** \name Import and Export Methods */
	//@{				  
				  
	 /** \brief Exports the DAG to a gl file.
	 \deprecated
	  *\param[in] filename the file name of the DDG in gl format
	  \remarks if multiple registers types exist in the architecture, then the output gl file will consider them as the same register type, because the current gl format does not support multiple register files.
	  */
	void write_to_gl(const char * filename) const;
	
	/** \brief Exports the DAG to a gl file. 
	\deprecated
	  *\param[in] filename the file name of the DDG in gl format
	  *\param[in] regtype_id An id of a register type to consider for the flow_reg dependances in the gl file. Only flow edges through registers of type regtype_id will be output as flow_reg. Other flow of other register types are considered as serial.
	  */
	void write_to_gl(const char * filename , int regtype_id) const;
	
  	
	/** \brief Imports  the DAG from  a gl file when a unique register type exists in the architecture.  
	\deprecated
	 * \param[in] filename the file name of the DDG in gl format
	  \returns true if OK. 
	\remarks 
  		- To get the DAG from a cyclic data dependence graph, all edges with non null distances are ignored. That is, only the loop body is considered when importing a cyclic data dependence graph from a gl file.
   	* 	- The set \f$ E_R \f$ of flow edges through registers is constructed for the unique register type in the architecture.
		-DAG::build_internal_structures() is called inside this method.

  	* \throws BadIODDG If an IO error occurs (corrupted or non present input files for instance), an exception object of type DDG::BadIODDG is thrown
	 * \pre Before reading a DAG from a gl file, it is assumed that the ISA of the DAG defines all the opcodes ids of the nodes as present in the gl file. Also, it is assumed that the ISA has a unique register type (this is a restriction of the gl format). The constructor DDG::DAG(ARCHITECTURE) defines a DAG with a user defined ISA, otherwise a default ISA is assumed, see  ISA_xml.
  	 */
  	bool read_from_gl(const char *filename);
  	
  	
  	/** \brief Imports  the DAG from  a gl file when multiple registers types exist in the architecture. This function asks to fix which register type to consider.  
	\deprecated
	 * \param[in] filename the file name of the DDG in gl format
	 * \param[in] regtype_id The id of the register type to consider for flow_reg dependances.
	 \returns true if OK. 
	\remarks 
  		- To get the DAG from a cyclic data dependence graph, all edges with non null distances are ignored. That is, only the loop body is considered when importing a cyclic data dependence graph from a gl file.
   	* 	- The set \f$ E_R \f$ of flow edges through registers is constructed for the register type given as parameter.
		-DAG::build_internal_structures() is called inside this method.

  	* \throws BadIODDG If an IO error occurs (corrupted or non present input files for instance), an exception object of type DDG::BadIODDG is thrown
	 * \pre Before reading a DAG from a gl file, it is assumed that the ISA of the DAG defines all the opcodes ids of the nodes as present in the gl file. The constructor DDG::DAG(ARCHITECTURE) defines a DAG with a user defined ISA, otherwise a default ISA is assumed, see  \ref ISA_xml.
  	 */
  	bool read_from_gl(const char *filename, int regtype_id);
		
	/** \brief Exports the DAG to a vcg file to in order to be visualized with the xvcg tool.  
	 * The user can navigate using vcg on all DDG attributes (types of edges, opcodes, etc.)
	\param[in] filename the file name of the DDG in vcg format
	 * For more information about xvcg, see http://rw4.cs.uni-sb.de/users/sander/html/gsvcg1.html
	*/
  	void write_to_vcg(const char *filename) const;

	/** \brief Exports the DAG to a vcg file to in order to be visualized with the xvcg tool while considering one register type.
	 * The user can navigate using vcg on all DDG attributes (types of edges, opcodes, etc.)
	 * For more information about xvcg, see http://rw4.cs.uni-sb.de/users/sander/html/gsvcg1.html
	\param[in] filename the file name of the DDG in vcg format
	\param[in] t the id of the considered register type
	 */
	void write_to_vcg(const char *filename, int t) const;
	
	/** \brief Exports the DAG to an XML description.
		
	 * All DDG attributes are included into this XML graph format.*/
	
	void write_to_xml(const char *filename);
	/** \brief Imports  the DAG from its xml description.
	 \returns True if OK. False if problem.
  	 */
	bool read_from_xml(const char *filename);
	
	
	
   

	
	
	//@}
    
   
    

};


 /**
 * \fn int MAXIMAL_ANTI_CHAIN(const graph & G, set<node>& MA);
 * @param[in] G An acyclic graph (DAG, order)
 * @param[out] MA The set of nodes belonging to a maximal antichain
 * @return The size of a maximal antichain in the graph (-1 if the graph is not acyclic). Returns 0 if the input acyclic graph is empty. Returns -1 if error.

 This function implements Dilworth decomposition to compute maximal antichain inside an order (DAG).
 * \remarks
* A maxi�al antichain may not be unique. This function returns an arbitrary one.
  \pre The input graph G should be acyclic.
 */
int MAXIMAL_ANTI_CHAIN(const graph & G, set<node>& MA);

/**
 * \fn int MINIMAL_CHAIN(const graph &G, node_array<int>& chain)
 * @param[in] G An acyclic graph (DAG, order)
 * @param[out] chain Contains the chain number for each node of the DAG. \f$ \forall u \in V, chain[u] \f$ is the number of the chain to which \f$ u \f$ belongs.
 * @return The size of a minimal chain decomposition of the graph. Returns 0 if the input acyclic graph is empty. Returns -1 if error.

 This function implements Dilworth decomposition to compute minimal chain decomposition of an order (DAG).
 \pre The input graph G should be acyclic.
 * \remarks
* A minimal chain decomposition may not be unique. This function returns an arbitrary one.
 */
int MINIMAL_CHAIN(const graph &G, node_array<int>& chain);

/** int MINIMAL_CHAIN(const DAG &G, array< list<node> > &C)
 * This function computes a minimal chain decomposition of a DAG (an order).
 * @param[in] G An input DAG
 * @param[out] C \f$ C[i] \f$ is a node list reprensenting the chain number \f$ i \f$, with \f$ 0 \le i < p(G) \f$. The lists are sorted in increasing order: \f$ \forall i, \forall u,v \in C[i], u \le v \Longleftrightarrow \exists \textrm{ a path } (u,\ldots,v) \textrm{ in }G \f$.
 * @return \f$ p(G) \f$  the minimal number of chains of the DAG. Dilworth proved that \f$  p(G)=w(G)\f$ , that is, it is equal to the size of a maximal antichain. Returns 0 if the input acyclic graph is empty. Returns -1 if error.
* \remarks
* A minimal chain decomposition may not be unique. This function returns an arbitrary one.
* \pre G is acyclic.

 */

int MINIMAL_CHAIN(const DAG &G, array< list<node> > &C);

/**
 * \fn int REGISTER_SATURATION (const DAG & G, leda::list<node> & RS_values, bool fast = true);
 * @param[in] G A DAG
 * @param[out] RS_values The list of saturating values. It is proved that there exists an instruction schedule for any underlying processor constraints that make all the saturating values simultaneously alive.
 * @param[in] fast boolean to indicate to use either fast heuristic or accurate one
 * @return An approximate register saturation for the DAG. 
 \pre The set \f$ V_R \subseteq V \f$ of values  and the set \f$ E_R \subseteq E \f$ of flow edges should be set.
  \pre a unique register type exists in the ISA.
 \throws NonUniqueRegisterType
 
 This function implements a heuristics for computing the register saturation of the DAG  (computing the optimal register saturation is an NP-complete problem). It consists of computing the maximal register need of a DAG independently of instruction schedules. Our experiments on a large set of DAGs show that our heuristics is optimal in 95\% of the cases, and far from the optimal by only one register in 5\% of the cases. This concept has been mathematically and experimentally validated in the follwing research article :
 
 Sid Touati. Register Saturation in Instruction Level Parallelism. International Journal of Parallel Programming, Springer-Verlag, Volume 33, Issue 4, August 2005.
 */
 int REGISTER_SATURATION (const DAG & G, leda::list<node> & RS_values, bool fast = true);


/**
 * \fn  int REGISTER_SATURATION (const DAG & G, leda::list<node> & RS_values, node_array<node> &k, bool fast = true);
 * @param[in] G An input DAG
 * @param[out] RS_values The list of saturating values. It is proved that there exists an instruction schedule for any underlying processor constraints that make all the saturating values simultaneously alive.
  @param[out] k A valid saturating killing function. If the computed register saturation is equal to RS, then it is proved that there exists a schedule for the input DAG (for any hardware property) that requires RS registers if the killer of each value \f$ u \in V_R \f$ is set to \f$ k(u) \f$.
 * @param[in] fast boolean to indicate to use either fast heuristic or accurate one
 * @return An approximate register saturation for the DAG. 
 \pre The set \f$ V_R \subseteq V \f$ of values  and the set \f$ E_R \subseteq E \f$ of flow edges should be set.
 \pre a unique register type exists in the ISA.
 \throws NonUniqueRegisterType
 This function implements a heuristics for computing the register saturation of the DAG  (computing the optimal register saturation is an NP-complete problem). It consists of computing the maximal register need of a DAG independently of instruction schedules. Our experiments on a large set of DAGs show that our heuristics is optimal in 95\% of the cases, and far from the optimal by only one register in 5\% of the cases. This concept has been mathematically and experimentally validated in the follwing research article :
 
 Sid Touati. Register Saturation in Instruction Level Parallelism. International Journal of Parallel Programming, Springer-Verlag, Volume 33, Issue 4, August 2005.
 */
 int REGISTER_SATURATION (const DAG & G, leda::list<node> & RS_values, node_array<node> &k, bool fast = true);



/**
 * \fn int REGISTER_SATURATION (const DAG & G, int t, leda::list<node> & RS_values, bool fast = true);
 * @param[in] G A DAG
 * @param[in] t A register type identifier.
 * @param[out] RS_values The list of saturating values of type t. It is proved that there exists an instruction schedule for any underlying processor constraints that make all the saturating values simultaneously alive.
 * @param[in] fast boolean to indicate to use either fast heuristic or accurate one
 * @return An approximate register saturation for the DAG. 
 \pre The set \f$ V_{R,t} \subseteq V \f$ of values  and the set \f$ E_{R,t} \subseteq E \f$ of flow edges should be set.
 This function implements a heuristics for computing the register saturation of the DAG  (computing the optimal register saturation is an NP-complete problem). It consists of computing the maximal register need of a DAG independently of instruction schedules. Our experiments on a large set of DAGs show that our heuristics is optimal in 95\% of the cases, and far from the optimal by only one register in 5\% of the cases. This concept has been mathematically and experimentally validated in the follwing research article :
 
 Sid Touati. Register Saturation in Instruction Level Parallelism. International Journal of Parallel Programming, Springer-Verlag, Volume 33, Issue 4, August 2005.
 */
 int REGISTER_SATURATION (const DAG & G, int t, leda::list<node> & RS_values, bool fast = true);


/**
 * \fn  int REGISTER_SATURATION (const DAG & G, int t, leda::list<node> & RS_values, node_array<node> &k, bool fast = true);
 * @param[in] G An input DAG
  * @param[in] t A register type identifier.
 * @param[out] RS_values The list of saturating values. It is proved that there exists an instruction schedule for any underlying processor constraints that make all the saturating values simultaneously alive.
  @param[out] k A valid saturating killing function. If the computed register saturation is equal to RS, then it is proved that there exists a schedule for the input DAG (for any hardware property) that requires RS registers if the killer of each value \f$ u \in V_{R,t} \f$ is set to \f$ k(u^t) \f$.
 * @param[in] fast boolean to indicate to use either fast heuristic or accurate one
 * @return An approximate register saturation for the DAG. 
 \pre The set \f$ V_{R,t} \subseteq V \f$ of values  and the set \f$ E_{R,t} \subseteq E \f$ of flow edges should be set.
 This function implements a heuristics for computing the register saturation of the DAG  (computing the optimal register saturation is an NP-complete problem). It consists of computing the maximal register need of a DAG independently of instruction schedules. Our experiments on a large set of DAGs show that our heuristics is optimal in 95\% of the cases, and far from the optimal by only one register in 5\% of the cases. This concept has been mathematically and experimentally validated in the follwing research article :
 
 Sid Touati. Register Saturation in Instruction Level Parallelism. International Journal of Parallel Programming, Springer-Verlag, Volume 33, Issue 4, August 2005.
 */
 int REGISTER_SATURATION (const DAG & G, int t, leda::list<node> & RS_values, node_array<node> &k, bool fast = true);

}


#endif
