 /***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

/*! \file RS.cpp
\brief This file contains the implementation of register saturation functions
This is an implementation of register saturation computation. Given a DAG, Computing the exact register saturation (maximal register pressure) is an NP-complete problem. In this file, we implement a heuristics as described in the research article :
Sid Touati. Register Saturation in Instruction Level Parallelism. International Journal of Parallel Programming, Springer-Verlag, Volume 33, Issue 4, August 2005.
*/


#include "RS.h"

namespace DDG{


  int REGISTER_SATURATION (const DAG & G, leda::list<node> & RS_values, bool fast){
  DAG pkg, dvg;
  node s;
  std::list<CBC> all_cb;
  node_map<node> original_node, dvg_node, pk_node;
  node_array<node> k(G); //saturating killing function
  set<node> MA; /* maximal antichain in the killing graph */
  CBC cb;
  int RS=0; //register saturation
  get_PKG(G, pkg, original_node, pk_node); // compute the potential killing DAG
  all_cb=override_decompose_into_cbc(pkg);  
  
  std::list<CBC>::iterator iter;
  for(iter = all_cb.begin(); iter != all_cb.end(); iter++){
 	cb = (CBC)*iter;
      SKS_greedy(G, pkg, original_node, pk_node, cb, k);
  }

  //
  if(fast) {
    get_DVG(G, k, dvg, original_node, dvg_node);
  } else {
    graph Gk;
    node_map <node> g2gk_node, gk2g_node;
    get_Gk(G, pkg, pk_node, original_node, k, Gk, g2gk_node, gk2g_node);
    get_DVG(G, k, Gk, g2gk_node, gk2g_node, dvg, original_node, dvg_node);
  }


  RS=MAXIMAL_ANTI_CHAIN(dvg, MA);
  RS_values.clear();
  forall(s, MA){
  	RS_values.push(original_node[s]);
  }
  return RS_values.size();
}

  int REGISTER_SATURATION (const DAG & G, leda::list<node> & RS_values, node_array<node> &k, bool fast){
  DAG pkg, dvg;
  node s;
  std::list<CBC> all_cb;
  node_map<node> original_node, dvg_node, pk_node;
  set<node> MA; /* maximal antichain in the killing graph */
  CBC cb;
  int RS=0; //register saturation
  k.init(G); //saturating killing function  
  get_PKG(G, pkg, original_node, pk_node); // compute the potential killing DAG
  all_cb=override_decompose_into_cbc(pkg);  
  
  std::list<CBC>::iterator iter;
  for(iter = all_cb.begin(); iter != all_cb.end(); iter++){
 	cb = (CBC)*iter;
      SKS_greedy(G, pkg, original_node, pk_node, cb, k);
  }
  
  if(fast) {
    get_DVG(G, k, dvg, original_node, dvg_node);
  } else {
    graph Gk;
    node_map <node> g2gk_node, gk2g_node;
    get_Gk(G, pkg, pk_node, original_node, k, Gk, g2gk_node, gk2g_node);
    get_DVG(G, k, Gk, g2gk_node, gk2g_node, dvg, original_node, dvg_node);
  }

  RS=MAXIMAL_ANTI_CHAIN(dvg, MA);
  RS_values.clear();
  forall(s, MA){
  	RS_values.push(original_node[s]);
  }
  return RS_values.size();

}

void SKS_greedy(const DAG &G, const DAG & PKG, const node_map<node> original_node, const node_map<node> pk_node, const CBC cb, node_array<node> &k){
	node s, t,u;
	set<node> X, Y;
	EN en;
	set<node> Gt, GammaMinus_t;
	leda::list<EN> T;
	node_array<bool> covered(G, false);

	forall(s, cb.S) {X.insert(original_node[s]);}

	/* initialize the list of children with their costs */
	forall(t, cb.T){
		en.t = original_node[t];
		T.append(en);
	}
	/* apply the greedy heuristic */
	while(!X.empty()){
		/* selct the child with the maximum cost */
		compute_children_cost(X, Y, T, G, PKG, original_node, pk_node);
		t=(T[T.max()]).t;
		/* compute GammaMinus_t */		
		Gt=PKG.get_parents(pk_node[t]);
		forall(s, Gt){
			GammaMinus_t.insert(original_node[s]);
		}		

		X-=GammaMinus_t;
		/* cumullated descnedent values */		
		Y+=(G.get_down(t) & G.get_V_R());
		
		/* kill and cover non killed parents */
		forall(s, GammaMinus_t){
			if(!covered[s]){
				covered[s]=true;
				k[s]=t;
			}
		}
	}
}

void SKS_greedy(const DAG &G, int regtype, const DAG & PKG, const node_map<node> original_node, const node_map<node> pk_node, const CBC cb, node_array<node> &k){
	node s, t,u;
	set<node> X, Y;
	EN en;
	set<node> Gt, GammaMinus_t;
	leda::list<EN> T;
	node_array<bool> covered(G, false);

	forall(s, cb.S) {X.insert(original_node[s]);}

	/* initialize the list of children with their costs */
	forall(t, cb.T){
		en.t = original_node[t];
		T.append(en);
	}
	/* apply the greedy heuristic */
	while(!X.empty()){
		/* selct the child with the maximum cost */
		compute_children_cost(regtype, X, Y, T, G, PKG, original_node, pk_node);
		t=(T[T.max()]).t;
		/* compute GammaMinus_t */		
		Gt=PKG.get_parents(pk_node[t]);
		forall(s, Gt){
			GammaMinus_t.insert(original_node[s]);
		}		

		X-=GammaMinus_t;
		/* cumullated descnedent values */		
		Y+=(G.get_down(t) & G.get_V_R(regtype));
		
		/* kill and cover non killed parents */
		forall(s, GammaMinus_t){
			if(!covered[s]){
				covered[s]=true;
				k[s]=t;
			}
		}
	}
}

rational rho(const DAG &G, const set<node> X, const set<node> Y, const node t, const set<node>  GammaMinus){
	int nomi, denomi, id;
	node u;
	rational cost;
	set<node> down_val_t=G.get_down(t), temp;
	down_val_t&=(G.get_V_R());

	temp=GammaMinus & X;	
	nomi=temp.size();
	
	temp=down_val_t + Y;
	
	denomi=temp.size();
// 	if(denomi!=0) cost= rational(nomi, denomi);
// 	else cost=rational(nomi, 1);
	// Sebastien Briais has found a better cost model like this
	
	cost= rational(nomi, 1+denomi);
	return cost;
}

rational rho(const DAG &G, int regtype, const set<node> X, const set<node> Y, const node t, const set<node>  GammaMinus){
	int nomi, denomi, id;
	node u;
	rational cost;
	set<node> down_val_t=G.get_down(t), temp;
	down_val_t&=(G.get_V_R(regtype));

	temp=GammaMinus & X;	
	nomi=temp.size();
	
	temp=down_val_t + Y;
	
	denomi=temp.size();
// 	if(denomi!=0) cost= rational(nomi, denomi);
// 	else cost=rational(nomi, 1);
	// Sebastien Briais has found a better cost model like this

	cost= rational(nomi, 1+denomi);
	return cost;
}



void compute_children_cost(const set<node> X, const set<node> Y, leda::list<EN> &T,  const DAG &G, const DAG &PKG, const node_map<node> original_node, const node_map<node> pk_node){
	EN en;
	list_item it;
	node u;
	set<node> GammaMinus, P;
	forall_items(it, T){
		en=T[it];
		GammaMinus.clear();
		P=PKG.get_parents(pk_node[en.t]);
		forall(u,P){
			GammaMinus.insert(original_node[u]);
		}
		en.rho=rho(G, X, Y, en.t, GammaMinus);
		T[it]=en;
	}
}

void compute_children_cost(int regtype, const set<node> X, const set<node> Y, leda::list<EN> &T,  const DAG &G, const DAG &PKG, const node_map<node> original_node, const node_map<node> pk_node){
	EN en;
	list_item it;
	node u;
	set<node> GammaMinus, P;
	forall_items(it, T){
		en=T[it];
		GammaMinus.clear();
		P=PKG.get_parents(pk_node[en.t]);
		forall(u,P){
			GammaMinus.insert(original_node[u]);
		}
		en.rho=rho(G, regtype, X, Y, en.t, GammaMinus);
		T[it]=en;
	}
}

// this function has been implemented vby Frederic Brault using the STL lists instead of LEDA lists.
// Indeed, LEDA 5.2 lists are buggy on 64 bits.
std::list<CBC> override_decompose_into_cbc(const DAG &G){
	std::list<CBC> all_cb;
	CBC cb;
	set<node> S, T, tmp;
	node_array<bool> visited (G, false);
	node s, v, sp, t, tp;	
	forall_nodes(s, G){
		if(!visited[s] && (G.outdeg(s)>0)){
			cb.clear();
			visited[s]=true;
			cb.S.insert(s);
			cb.T+=G.get_children(s);
			S.clear();T.clear();
			/* build S and T */
			while((S!=cb.S) || (T!=cb.T)){
				S=cb.S;
				T=cb.T;
				forall(tp, cb.T){cb.S+=G.get_parents(tp);}
				forall(sp, cb.S){cb.T+=G.get_children(sp);}
			}
			/* build E_CB */
			tmp.clear();
			forall(s, S){
				visited[s]=true;
				tmp.insert(s);
				cb.E_CB+=list_to_set(G.out_edges(s));
			}
			cb.S=tmp;
			tmp.clear();
			forall(s, T){
				tmp.insert(s);
			}
			cb.T=tmp;
			tmp.clear();
			all_cb.push_back(cb);
		}// end if
	}

	return all_cb;
}


leda::list<CBC> decompose_into_cbc(const DAG &G){
	leda::list<CBC> all_cb;
	CBC cb;
	set<node> S, T, tmp;
	node_array<bool> visited (G, false);
	node s, v, sp, t, tp;	
	forall_nodes(s, G){
		if(!visited[s] && (G.outdeg(s)>0)){
			cb.clear();
			visited[s]=true;
			cb.S.insert(s);
			cb.T+=G.get_children(s);
			S.clear();T.clear();
			/* build S and T */
			while((S!=cb.S) || (T!=cb.T)){
				S=cb.S;
				T=cb.T;
				forall(tp, cb.T){cb.S+=G.get_parents(tp);}
				forall(sp, cb.S){cb.T+=G.get_children(sp);}
			}
			/* build E_CB */
			tmp.clear();
			forall(s, S){
				visited[s]=true;
				tmp.insert(s);
				cb.E_CB+=list_to_set(G.out_edges(s));
			}
			cb.S=tmp;
			tmp.clear();
			forall(s, T){
				tmp.insert(s);
			}
			cb.T=tmp;
			tmp.clear();
			all_cb.append(cb);
		}// end if
	}

	return all_cb;
}

void get_PKG(const DAG &G, DAG &PKG, node_map<node> &original_node, node_map<node> &pk_node){
  node u, v, n;
  edge e;
  INFO_EDGE ie=INFO_EDGE(killerdep);
  set<node> S;
  INFO_NODE in;
  int id;
  set<node> pkill_u;
  node_array<set<node> > pkill; /* potential killing set for each node */
  node_array<set<node> > Cons; /* consumers of each value */
  pk_node.init(G);
  PKG.clear();
  PKG.set_ISA(G.get_ISA());
  // duplicates the node in PKG
  forall_nodes(u,G) {
  	id=PKG.new_node((G.instruction_type(u)).opcode_id());
  	pk_node[u]=PKG.node_with_id(id);  // save the mapping G -> PKG

  }
  original_node.init(PKG);
  forall_nodes(v,PKG) {
  	forall_nodes(u, G){
  	 if( pk_node[u]==v) {
  	  original_node[v]=u; //save the mapping PKG -> G
  	  break;
  	  }
  	}
  }
  
  /* build Consumer set of each value */
  Cons.init(G);
  forall_nodes(u, G){
    if(G.is_value(u)){
      forall_out_edges(e, u){
	if(G.is_flow_reg(e)){
	  Cons[u].insert(G.target(e));
	}
      }
    }
  }

  /* compute the potential killing set of each node */
  pkill.init(G);
    forall_nodes(u, G){
    if(G.is_value(u)){
      pkill[u]=Cons[u];
      S=Cons[u];
      forall(v, S){
	forall(n, S){
	  if((v!=n)&&((G.get_down(n)).member(v))) pkill[u].del(n);
	}
      }
    }
  }

  /* insert the pkill edges in PKG */
  forall_nodes(u, PKG){
      if (PKG.is_value(u)){
      pkill_u = pkill[original_node[u]];
      forall(v, pkill_u){
	ie.set_delta(PKG.latency(u));
	PKG.new_edge(u,pk_node[v], ie);
      }
      }
  }
  PKG.build_internal_structures();
}

void get_PKG(const DAG &G, int t, DAG &PKG, node_map<node> &original_node, node_map<node> &pk_node){
  node u, v, n;
  edge e;
  INFO_EDGE ie=INFO_EDGE(killerdep);
  set<node> S;
  INFO_NODE in;
  int id;
  set<node> pkill_u;
  node_array<set<node> > pkill; /* potential killing set for each node */
  node_array<set<node> > Cons; /* consumers of each value */
  pk_node.init(G);
  PKG.clear();
  PKG.set_ISA(G.get_ISA());
  // duplicates the node in PKG
  forall_nodes(u,G) {
  	id=PKG.new_node((G.instruction_type(u)).opcode_id());
  	pk_node[u]=PKG.node_with_id(id);  // save the mapping G -> PKG

  }
  original_node.init(PKG);
  forall_nodes(v,PKG) {
  	forall_nodes(u, G){
  	 if( pk_node[u]==v) {
  	  original_node[v]=u; //save the mapping PKG -> G
  	  break;
  	  }
  	}
  }
  
  /* build Consumer set of each value */
  Cons.init(G);
  forall_nodes(u, G){
    if(G.is_value(u,t)){
      forall_out_edges(e, u){
	if(G.is_flow_reg(e,t)){
	  Cons[u].insert(G.target(e));
	}
      }
    }
  }

  /* compute the potential killing set of each node */
  pkill.init(G);
    forall_nodes(u, G){
    if(G.is_value(u,t)){
      pkill[u]=Cons[u];
      S=Cons[u];
      forall(v, S){
	forall(n, S){
	  if((v!=n)&&((G.get_down(n)).member(v))) pkill[u].del(n);
	}
      }
    }
  }

  /* insert the pkill edges in PKG */
  forall_nodes(u, PKG){
      if (PKG.is_value(u,t)){
      pkill_u = pkill[original_node[u]];
      forall(v, pkill_u){
	ie.set_delta(PKG.latency(u));
	PKG.new_edge(u,pk_node[v], ie);
      }
      }
  }
  PKG.build_internal_structures();
}

  //
  void get_Gk(const DAG &G, const DAG &PKG, const node_map<node> &g2pkg_node, const node_map<node> &pkg2g_node, const node_array<node> k, 
	      graph &Gk, node_map<node> &g2gk_node, node_map<node> &gk2g_node) {
    Gk.clear();
    Gk.make_directed(); // is it necessary?

    g2gk_node.init(G);
    gk2g_node.init(Gk);

    // copy the structure of G
    {
      node u;
      forall_nodes(u, G) {
	node uk = Gk.new_node();
	g2gk_node[u] = uk;
	gk2g_node[uk] = u;
      }
    }

    {
      node u;
      forall_nodes(u, G) {
	node uk = g2gk_node[u];
	node v;
	forall_adj_nodes(v, u) {
	  // iterate over successors of u
	  node vk = g2gk_node[v];
	  Gk.new_edge(uk, vk);
	}
      }
    }

    // add new edges (v,k(u)) for each value u and v in (pkill(u) - k(u))
    {
      node u;
      set<node> VR = G.get_V_R();
      forall(u, VR) {
	node ku = k[u];
	if(ku == NULL) {
	  // ku = NULL means that there are no potential killers
	  continue;
	}
	node kuk = g2gk_node[ku];

	node u_pk = g2pkg_node[u];
	node v_pk;
	forall_adj_nodes(v_pk, u_pk) {
	  node v = pkg2g_node[v_pk];
	  if(v == ku) {
	    // skip killer of u
	    continue;
	  }
	  node vk = g2gk_node[v];
	  Gk.new_edge(vk, kuk);
	}
      }
    }

    MAKE_TRANSITIVELY_CLOSED(Gk);
  }

  void get_Gk(const DAG &G, int t, const DAG &PKG, const node_map<node> &g2pkg_node, const node_map<node> &pkg2g_node, const node_array<node> k, 
	      graph &Gk, node_map<node> &g2gk_node, node_map<node> &gk2g_node) {
    Gk.clear();
    Gk.make_directed(); // is it necessary?

    g2gk_node.init(G);
    gk2g_node.init(Gk);

    // copy the structure of G
    {
      node u;
      forall_nodes(u, G) {
	node uk = Gk.new_node();
	g2gk_node[u] = uk;
	gk2g_node[uk] = u;
      }
    }

    {
      node u;
      forall_nodes(u, G) {
	node uk = g2gk_node[u];
	node v;
	forall_adj_nodes(v, u) {
	  // iterate over successors of u
	  node vk = g2gk_node[v];
	  Gk.new_edge(uk, vk);
	}
      }
    }

    // add new edges (v,k(u)) for each value u and v in (pkill(u) - k(u))
    {
      node u;
      set<node> VR = G.get_V_R(t);
      forall(u, VR) {
	node ku = k[u];
	if(ku == NULL) {
	  // ku = NULL means that there are no potential killers
	  continue;
	}
	node kuk = g2gk_node[ku];

	node u_pk = g2pkg_node[u];
	node v_pk;
	forall_adj_nodes(v_pk, u_pk) {
	  node v = pkg2g_node[v_pk];
	  if(v == ku) {
	    // skip killer of u
	    continue;
	  }
	  node vk = g2gk_node[v];
	  Gk.new_edge(vk, kuk);
	}
      }
    }

    MAKE_TRANSITIVELY_CLOSED(Gk);
  }
  //

  void get_DVG(const DAG &G, const node_array<node> k, const graph &Gk, const node_map<node> &g2gk_node, const node_map<node> &gk2g_node, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node){
    node u,v;
    edge e;
    INFO_EDGE ie;
    set<node> VR=G.get_V_R();
    INFO_NODE in;
    int id;
    set<node> downval;
    
    dvg_node.init(G);
    DVG.clear();
    DVG.set_ISA(G.get_ISA());
    // duplicates the node in DVG
    forall(u,VR) {
      id=DVG.new_node((G.instruction_type(u)).opcode_id());
      dvg_node[u]=DVG.node_with_id(id);  // save the mapping G -> DVG
    }
    original_node.init(DVG);
    forall_nodes(v,DVG) {
      forall_nodes(u, G){
	if( dvg_node[u]==v) {
  	  original_node[v]=u; //save the mapping DVG -> G
  	  break;
	}
      }
    }
    
    {
      node dvu;
      forall_nodes(dvu, DVG) {
	node u = original_node[dvu];
	node ku = k[u];
	if(ku == NULL) {
	  continue;
	}
	node kuk = g2gk_node[ku];
	node vk;
	forall_adj_nodes(vk, kuk) {
	  node v = gk2g_node[vk];
	  if(G.is_value(v)) {
	    node dvv = dvg_node[v];
	    ie.set_delta(DVG.latency(dvu));
	    DVG.new_edge(dvu, dvv, ie);
	  }
	}
      }
    }

    MAKE_TRANSITIVELY_REDUCED(DVG);

    DVG.build_internal_structures();
  }

void get_DVG(const DAG &G, int t, const node_array<node> k, const graph &Gk, const node_map<node> &g2gk_node, const node_map<node> &gk2g_node, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node){
  node u,v;
  edge e;
  INFO_EDGE ie;
  set<node> VR=G.get_V_R(t);
  INFO_NODE in;
  int id;
  set<node> downval;

  dvg_node.init(G);
  DVG.clear();
  DVG.set_ISA(G.get_ISA());
  // duplicates the node in DVG
  forall(u,VR) {
  	id=DVG.new_node((G.instruction_type(u)).opcode_id());
  	dvg_node[u]=DVG.node_with_id(id);  // save the mapping G -> DVG
  }
  original_node.init(DVG);
  forall_nodes(v,DVG) {
  	forall_nodes(u, G){
  	 if( dvg_node[u]==v) {
  	  original_node[v]=u; //save the mapping DVG -> G
  	  break;
  	  }
  	}
  }

  {
    node dvu;
    forall_nodes(dvu, DVG) {
      node u = original_node[dvu];
      node ku = k[u];
      if(ku == NULL) {
	continue;
      }
      node kuk = g2gk_node[ku];
      node vk;
      forall_adj_nodes(vk, kuk) {
	node v = gk2g_node[vk];
	if(G.is_value(v,t)) {
	  node dvv = dvg_node[v];
	  ie.set_delta(DVG.latency(dvu));
	  DVG.new_edge(dvu, dvv, ie);
	}
      }
    }
  }

  MAKE_TRANSITIVELY_REDUCED(DVG);

  DVG.build_internal_structures();
}

void get_DVG(const DAG &G, const node_array<node> k, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node){
  node u,v;
  edge e;
  INFO_EDGE ie;
  set<node> VR=G.get_V_R();
  INFO_NODE in;
  int id;
  set<node> downval;

  dvg_node.init(G);
  DVG.clear();
  DVG.set_ISA(G.get_ISA());
  // duplicates the node in DVG
  forall(u,VR) {
  	id=DVG.new_node((G.instruction_type(u)).opcode_id());
  	dvg_node[u]=DVG.node_with_id(id);  // save the mapping G -> DVG
  }
  original_node.init(DVG);
  forall_nodes(v,DVG) {
  	forall_nodes(u, G){
  	 if( dvg_node[u]==v) {
  	  original_node[v]=u; //save the mapping DVG -> G
  	  break;
  	  }
  	}
  }


  forall_nodes(u, DVG){
    if(DVG.is_value(u)){
      if((v=k[original_node[u]])!=NULL) downval=G.get_down(v);
      else downval.clear();
      
      forall(v, downval){
	      ie.set_delta(DVG.latency(u));
		if( dvg_node[v] != NULL) DVG.new_edge(u,dvg_node[v], ie); // if dvg_node[v]==NULL => value killed outside the DAG = exit value
      }
    } 
  }
  
  MAKE_TRANSITIVELY_REDUCED(DVG);

  DVG.build_internal_structures();
}

void get_DVG(const DAG &G, int t, const node_array<node> k, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node){
  node u,v;
  edge e;
  INFO_EDGE ie;
  set<node> VR=G.get_V_R(t);
  INFO_NODE in;
  int id;
  set<node> downval;

  dvg_node.init(G);
  DVG.clear();
  DVG.set_ISA(G.get_ISA());
  // duplicates the node in DVG
  forall(u,VR) {
  	id=DVG.new_node((G.instruction_type(u)).opcode_id());
  	dvg_node[u]=DVG.node_with_id(id);  // save the mapping G -> DVG
  }
  original_node.init(DVG);
  forall_nodes(v,DVG) {
  	forall_nodes(u, G){
  	 if( dvg_node[u]==v) {
  	  original_node[v]=u; //save the mapping DVG -> G
  	  break;
  	  }
  	}
  }


  forall_nodes(u, DVG){
    if(DVG.is_value(u,t)){
      if((v=k[original_node[u]])!=NULL) downval=G.get_down(v);
      else downval.clear();
      
      forall(v, downval){
	      ie.set_delta(DVG.latency(u));
		if( dvg_node[v] != NULL) DVG.new_edge(u,dvg_node[v], ie); // if dvg_node[v]==NULL => value killed outside the DAG = exit value
      }
    } 
  }
  
  MAKE_TRANSITIVELY_REDUCED(DVG);

  DVG.build_internal_structures();
}


  int REGISTER_SATURATION (const DAG & G, int t, leda::list<node> & RS_values, bool fast){

  DAG pkg, dvg;
  node s;
  std::list<CBC> all_cb;
  node_map<node> original_node, dvg_node, pk_node;
  node_array<node> k(G); //saturating killing function
  set<node> MA; /* maximal antichain in the killing graph */
  CBC cb;
  int RS=0; //register saturation
  get_PKG(G, t, pkg, original_node, pk_node); // compute the potential killing DAG
  all_cb=override_decompose_into_cbc(pkg);  
  
  std::list<CBC>::iterator iter;
  for(iter = all_cb.begin(); iter != all_cb.end(); iter++){
 	cb = (CBC)*iter;
      SKS_greedy(G, t, pkg, original_node, pk_node, cb, k);
  }
  
  if(fast) {
    get_DVG(G, t, k, dvg, original_node, dvg_node);
  } else {
    graph Gk;
    node_map <node> g2gk_node, gk2g_node;
    get_Gk(G, t, pkg, pk_node, original_node, k, Gk, g2gk_node, gk2g_node);
    get_DVG(G, t, k, Gk, g2gk_node, gk2g_node, dvg, original_node, dvg_node);
  }

  RS=MAXIMAL_ANTI_CHAIN(dvg, MA);
  RS_values.clear();
  forall(s, MA){
  	RS_values.push(original_node[s]);
  }
  return RS_values.size();
}

  int REGISTER_SATURATION (const DAG & G, int t, leda::list<node> & RS_values, node_array<node> &k, bool fast){
  DAG pkg, dvg;
  node s;
  std::list<CBC> all_cb;
  node_map<node> original_node, dvg_node, pk_node;
  set<node> MA; /* maximal antichain in the killing graph */
  CBC cb;
  int RS=0; //register saturation
  k.init(G); //saturating killing function  
  get_PKG(G, t, pkg, original_node, pk_node); // compute the potential killing DAG
  all_cb=override_decompose_into_cbc(pkg);  
  
  std::list<CBC>::iterator iter;
  for(iter = all_cb.begin(); iter != all_cb.end(); iter++){
 	cb = (CBC)*iter;
      SKS_greedy(G, t, pkg, original_node, pk_node, cb, k);
  }

  if(fast) {
    get_DVG(G, t, k, dvg, original_node, dvg_node);
  } else {
    graph Gk;
    node_map <node> g2gk_node, gk2g_node;
    get_Gk(G, t, pkg, pk_node, original_node, k, Gk, g2gk_node, gk2g_node);
    get_DVG(G, t, k, Gk, g2gk_node, gk2g_node, dvg, original_node, dvg_node);
  }

  RS=MAXIMAL_ANTI_CHAIN(dvg, MA);
  RS_values.clear();
  forall(s, MA){
  	RS_values.push(original_node[s]);
  }
  return RS_values.size();



}
}



namespace leda {
/** \internal
* \fn  int compare(const EN& u1, const EN& u2);
* \brief Compare two DDG::EN objects (used for LEDA sorting algorithms).
*  \returns The sign of (u1.rho - u2.rho)
*/
int compare(const DDG::EN& u1, const DDG::EN& u2) {
  if (u1.rho < u2.rho) return -1;
  else if (u1.rho > u2.rho) return 1;
  else return 0;
}
}

