/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

#ifndef __RS_H
#define __RS_H

/*! \file RS.h
\brief This is the header file for the implementation of register saturation for a DAG.
Register Saturation is the fruit of deep fundamental research results. The full theory is described in the following article:
Sid Touati. Register Saturation in Instruction Level Parallelism. International Journal of Parallel Programming, Springer-Verlag, Volume 33, Issue 4, August 2005. 57 pages.
*/
namespace DDG{
class EN;
}

namespace leda {
/** \internal
* \fn  int compare(const EN& u1, const EN& u2);
* \brief Compare two DDG::EN objects (used for LEDA sorting algorithms).
*  \returns The sign of (u1.rho - u2.rho)
*/
int compare(const DDG::EN& u1, const DDG::EN& u2);
}


#include "DDG.h"
#include <list>
using namespace std;
using namespace leda;
namespace DDG{	
/** \internal
 * \class CBC
 * \brief bipartite component class, used internally by RS computation
*/
class CBC{
	public:
		set<node> S, T;
		set<edge> E_CB;
		CBC(): S(), T(), E_CB(){}
		void clear(){T.clear();S.clear(); E_CB.clear();}
		friend ostream& operator<<(ostream &os, const CBC & x){}
		friend istream& operator>>(istream &is, CBC& x) {}
};

/** \internal
 * \class EN
 * \brief A simple class used for sorting nodes according to their costs \f$ \rho_{X,Y} \f$.
 */
class EN{ 
	public:
		node t;
		rational rho;
		friend istream& operator>>(istream &is, EN& x){
			is>>x.t>>x.rho;
		}
		friend ostream& operator<<(ostream &os, const EN & x){
			os<<"("<<x.t<<" , "<<x.rho<<")";
		}
};

/** \internal
 * \fn LEDA::list<CBC> decompose_into_cbc(const DAG &PKG);
 * \brief This function takes a DAG (a potential killing DAG in practice) and decompose it into connected bipartie component
 * \returns A LEDA list of CBC objects (connected bipartite components)
* \param[in] PKG The potential killing DAG.
*/
LEDA::list<CBC> decompose_into_cbc(const DAG &PKG);

/** \internal
 * \fn LEDA::list<CBC> decompose_into_cbc(const DAG &PKG);
 * \brief This function takes a DAG (a potential killing DAG in practice) and decompose it into connected bipartie component
 * \returns An STL list of CBC objects (connected bipartite components)
* \param[in] PKG The potential killing DAG.
\remarks This function has been implemented using STL lists instead LEDA lists because LEDA 5.2 lists are buggy on 64 bits machines.
*/
std::list<CBC> override_decompose_into_cbc(const DAG &G);

/** \internal
* \fn  void SKS_greedy(const DAG &G, const DAG & PKG, const node_map<node> original_node, const node_map<node> pk_node, const CBC cb,  node_array<node> &k);
* \brief A greedy algorithm for computing a saturating killing set inside a bipartite component
* \param[in] G The original DAG
* \param[in] PKG The potential killing DAG
  \param[in] original_node is a mapping from the nodes of PKG to the original nodes of G. \f$ \forall u \in PKG, original_node[u]\f$ corresponds to the original node in G.
  \param[in] pk_node is a mapping from the nodes of initial DAG G to the PKG. \f$ \forall u \in G, pk_node[u]\f$ corresponds to the copy node in PKG.
* \param[in] cb The considered bipartite component
 * \param[out] k The updated killing function that associates a killer to each value (bolonging to the parents of the CBC)
 \pre A unique register type exists in the ISA
\throws NonUniqueRegisterType

*/
void SKS_greedy(const DAG &G, const DAG & PKG, const node_map<node> original_node, const node_map<node> pk_node, const CBC cb,  node_array<node> &k);


/** \internal
* \fn  void SKS_greedy(const DAG &G, int t,  const DAG & PKG, const node_map<node> original_node, const node_map<node> pk_node, const CBC cb,  node_array<node> &k);
* \brief A greedy algorithm for computing a saturating killing set of type t inside a bipartite component
* \param[in] G The original DAG
* \param[in] t A register type
* \param[in] PKG The potential killing DAG
  \param[in] original_node is a mapping from the nodes of PKG to the original nodes of G. \f$ \forall u \in PKG, original_node[u]\f$ corresponds to the original node in G.
  \param[in] pk_node is a mapping from the nodes of initial DAG G to the PKG. \f$ \forall u \in G, pk_node[u]\f$ corresponds to the copy node in PKG.
* \param[in] cb The considered bipartite component
 * \param[out] k The updated killing function that associates a killer to each value (bolonging to the parents of the CBC)
*/
void SKS_greedy(const DAG &G, int t, const DAG & PKG, const node_map<node> original_node, const node_map<node> pk_node, const CBC cb,  node_array<node> &k);


/** \internal
* \fn  rational rho(DAG &G, set<node> X, set<node> Y, node t, set<node>  GammaMinus_t);
* \brief Computes the cost \f$ \rho_{X,Y}(t) \f$ of a node \f$ t \f$.
*  \returns The cost \f$ \rho_{X,Y}(t) \f$.
\param[in] G The original DAG
\param[in] X The set \f$ X \f$ of non (yet) killed parents (belongs to G)
\param[in] Y The set \f$ Y \f$ of cumulated descendent values (belongs to G)
\param[in] t the considered node (belongs to G)
\param[in] GammaMinus Set of t's parents in a CBC (nodes belong to G)
\pre A unique register type exists in the ISA
\throws NonUniqueRegisterType
*/
rational rho(const DAG &G, const set<node> X, const set<node> Y, const node t, const set<node>  GammaMinus_t);



/** \internal
* \fn  rational rho(DAG &G, int regtype, set<node> X, set<node> Y, node t, set<node>  GammaMinus_t);
* \brief Computes the cost \f$ \rho_{X,Y}(t) \f$ of a node \f$ t \f$.
*  \returns The cost \f$ \rho_{X,Y}(t) \f$.
\param[in] G The original DAG
\param[in] regtype A register type identifier
\param[in] X The set \f$ X \f$ of non (yet) killed parents (belongs to G)
\param[in] Y The set \f$ Y \f$ of cumulated descendent values (belongs to G)
\param[in] t the considered node (belongs to G)
\param[in] GammaMinus Set of t's parents in a CBC (nodes belong to G)
*/
rational rho(const DAG &G, int regtype, const set<node> X, const set<node> Y, const node t, const set<node>  GammaMinus);

/** \internal
* \fn  void compute_children_cost(set<node> X, set<node> Y, leda::list<EN> &T,  DAG &G, DAG &PKG, node_map<node> original_node, node_map<node> pk_node);

* \brief Computes the cost \f$ \rho_{X,Y}(t) \f$ of a list of children in a bipartite component.
\param[in] X The set \f$ X \f$ of non (yet) killed parents (belongs to G)
\param[in] Y The set \f$ Y \f$ of cumulated descendent values (belongs to G)
\param[out] T the list of costs (DDG:EN objects) of the considered chlidren in T
\param[in] G The original DAG
\param[in] PKG The potential killing DAG
\param[in] original_node is a mapping from the nodes of PKG to the original nodes of G. \f$ \forall u \in PKG, original_node[u]\f$ corresponds to the original node in G.
\param[in] pk_node is a mapping from the nodes of initial DAG G to the PKG. \f$ \forall u \in G, pk_node[u]\f$ corresponds to the copy node in PKG.
\pre A unique register type exists in the ISA
\throws NonUniqueRegisterType
*/
void compute_children_cost(const set<node> X, const set<node> Y, leda::list<EN> &T,  const DAG &G, const DAG &PKG, const node_map<node> original_node, const node_map<node> pk_node);

/** \internal
* \fn  void compute_children_cost(int regtype, set<node> X, set<node> Y, leda::list<EN> &T,  DAG &G, DAG &PKG, node_map<node> original_node, node_map<node> pk_node);
* \brief Computes the cost \f$ \rho_{X,Y}(t) \f$ of a list of children in a bipartite component.
\param[in] regtype A register type id
\param[in] X The set \f$ X \f$ of non (yet) killed parents (belongs to G)
\param[in] Y The set \f$ Y \f$ of cumulated descendent values (belongs to G)
\param[out] T the list of costs (DDG:EN objects) of the considered chlidren in T
\param[in] G The original DAG
\param[in] PKG The potential killing DAG
\param[in] original_node is a mapping from the nodes of PKG to the original nodes of G. \f$ \forall u \in PKG, original_node[u]\f$ corresponds to the original node in G.
\param[in] pk_node is a mapping from the nodes of initial DAG G to the PKG. \f$ \forall u \in G, pk_node[u]\f$ corresponds to the copy node in PKG.

*/
void compute_children_cost(int regtype, const set<node> X, const set<node> Y, leda::list<EN> &T,  const DAG &G, const DAG &PKG, const node_map<node> original_node, const node_map<node> pk_node);

/** \internal
* \fn  void get_PKG(DAG &G, DAG &PKG, node_map<node> &original_node, node_map<node> &pk_node,);
* \brief Computes potential killing DAG
\f$ \forall e=(u,v)\in PKG, v\in pkill(u). \f$
\param[in] G the initial DAG
\param[out] PKG The computed potential killing DAG
\param[out] original_node is a mapping from the nodes of PKG to the original nodes of G. \f$ \forall u \in PKG, original_node[u]\f$ corresponds to the original node in G.
\param[out] pk_node is a mapping from the nodes of initial DAG G to the PKG. \f$ \forall u \in G, pk_node[u]\f$ corresponds to the copy node in PKG.
\pre A unique register type exists in the ISA
\throws NonUniqueRegisterType
*/
void get_PKG(const DAG &G, DAG &PKG, node_map<node> &original_node, node_map<node> &pk_node);

/** \internal
* \fn  void get_PKG(DAG &G, int t,  DAG &PKG, node_map<node> &original_node, node_map<node> &pk_node,);
* \brief Computes potential killing DAG
\f$ \forall e=(u,v)\in PKG, v\in pkill(u). \f$
\param[in] t A register type id
\param[in] G the initial DAG
\param[out] PKG The computed potential killing DAG
\param[out] original_node is a mapping from the nodes of PKG to the original nodes of G. \f$ \forall u \in PKG, original_node[u]\f$ corresponds to the original node in G.
\param[out] pk_node is a mapping from the nodes of initial DAG G to the PKG. \f$ \forall u \in G, pk_node[u]\f$ corresponds to the copy node in PKG.
*/
void get_PKG(const DAG &G, int t, DAG &PKG, node_map<node> &original_node, node_map<node> &pk_node);

/** \internal
* \fn  void DAG::get_DVG(const DAG &G, const node_array<node> kill, const graph &Gk, const node_map<node> &g2gk_node, const node_map<node> &gk2g_node, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node);
* \brief Computes disjoint value DAG given a killing function
\param[in] G the initial DAG
\param[in] kill A valid killing function associating a killer to each value
\param[in] Gk the transitive closure of G_k order (G_k is the extended DAG associated with k)
\param[in] g2gk_node is a mapping from the nodes of G to the nodes of Gk
\param[in] gk2g_node is a mapping from the nodes of Gk to the nodes of G
\param[out] DVG The computed dispoint value DAG
\param[out] original_node is a mapping from the nodes of DVG to the original nodes of G. \f$ \forall u \in DVG, original_node[u]\f$ corresponds to the original node in G.
\param[out] dvg_node is a mapping from the nodes of initial DAG G to the DVG. \f$ \forall u \in G, dvg_node[u]\f$ corresponds to the copy node in DVG.
\pre A unique register type exists in the ISA
\throws NonUniqueRegisterType
*/
 void get_DVG(const DAG &G, const node_array<node> kill, const graph &Gk, const node_map<node> &g2gk_node, const node_map<node> &gk2g_node, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node);

/** \internal
* \fn  void DAG::get_DVG(const DAG &G, int t, const node_array<node> k, const graph &Gk, const node_map<node> &g2gk_node, const node_map<node> &gk2g_node, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node);
* \brief Computes disjoint value DAG given a killing function
\param[in] G the initial DAG
\param[in] t A register type id
\param[in] kill A valid killing function associating a killer to each value
\param[in] Gk the transitive closure of G_k order (G_k is the extended DAG associated with k)
\param[in] g2gk_node is a mapping from the nodes of G to the nodes of Gk
\param[in] gk2g_node is a mapping from the nodes of Gk to the nodes of G
\param[out] DVG The computed dispoint value DAG
\param[out] original_node is a mapping from the nodes of DVG to the original nodes of G. \f$ \forall u \in DVG, original_node[u]\f$ corresponds to the original node in G.
\param[out] dvg_node is a mapping from the nodes of initial DAG G to the DVG. \f$ \forall u \in G, dvg_node[u]\f$ corresponds to the copy node in DVG.
*/
 void get_DVG(const DAG &G, int t, const node_array<node> k, const graph &Gk, const node_map<node> &g2gk_node, const node_map<node> &gk2g_node, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node);

 /** \internal
  * \fn  void DAG::get_DVG(const DAG &G, int t, const node_array<node> k, const graph &Gk, const node_map<node> &g2gk_node, const node_map<node> &gk2g_node, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node);
  * \brief Computes disjoint value DAG given a killing function
 \param[in] G the initial DAG
 \param[in] t A register type id
 \param[in] kill A valid killing function associating a killer to each value
 \param[out] DVG The computed dispoint value DAG
 \param[out] original_node is a mapping from the nodes of DVG to the original nodes of G. \f$ \forall u \in DVG, original_node[u]\f$ corresponds to the original node in G.
 \param[out] dvg_node is a mapping from the nodes of initial DAG G to the DVG. \f$ \forall u \in G, dvg_node[u]\f$ corresponds to the copy node in DVG.
  */
 void get_DVG(const DAG &G, int t, const node_array<node> k,  DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node);
 
 
/** \internal
  * \fn  void DAG::get_DVG(const DAG &G, int t, const node_array<node> k, const graph &Gk, const node_map<node> &g2gk_node, const node_map<node> &gk2g_node, DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node);
  * \brief Computes disjoint value DAG given a killing function
 \param[in] G the initial DAG
 \param[in] kill A valid killing function associating a killer to each value
 \param[out] DVG The computed dispoint value DAG
 \param[out] original_node is a mapping from the nodes of DVG to the original nodes of G. \f$ \forall u \in DVG, original_node[u]\f$ corresponds to the original node in G.
 \param[out] dvg_node is a mapping from the nodes of initial DAG G to the DVG. \f$ \forall u \in G, dvg_node[u]\f$ corresponds to the copy node in DVG.
 */
 void get_DVG(const DAG &G, const node_array<node> k,  DAG &DVG, node_map<node> &original_node, node_map<node> &dvg_node);
 
 
 /** \internal
  * \fn  void DAG::get_Gk(const DAG &G, const DAG &PKG, const node_map<node> &g2pkg_node, const node_map<node> &pkg2g_node, const node_array<node> k, graph &Gk, node_map<node> &g2gk_node, node_map<node> &gk2g_node);
  * \brief Computes the transitive closure of the order induced by G_k (which is G extended by k)
  \param[in] G the initial DAG
  \param[in] PKG the potential killers DAG
  \param[in] g2pkg_node is a mapping from the nodes of G to the nodes of PKG
  \param[in] pkg2g_node is a mapping from the nodes of PKG to the nodes of G
  \param[in] k is a valid killing function
  \param[out] Gk is the computed graph
  \param[out] g2gk_node is a mapping from the nodes of G to the nodes of Gk
  \param[out] gk2g_node is a mapping from the nodes of Gk to the nodes of G
 */
 void get_Gk(const DAG &G, const DAG &PKG, const node_map<node> &g2pkg_node, const node_map<node> &pkg2g_node, const node_array<node> k, graph &Gk, node_map<node> &g2gk_node, node_map<node> &gk2g_node);
 
 /** \internal
  * \fn  void DAG::get_Gk(const DAG &G, int t, const DAG &PKG, const node_map<node> &g2pkg_node, const node_map<node> &pkg2g_node, const node_array<node> k, graph &Gk, node_map<node> &g2gk_node, node_map<node> &gk2g_node);
  \param[in] G the initial DAG
  \param[in] t A register type id
  \param[in] PKG the potential killers DAG
  \param[in] g2pkg_node is a mapping from the nodes of G to the nodes of PKG
  \param[in] pkg2g_node is a mapping from the nodes of PKG to the nodes of G
  \param[in] k is a valid killing function
  \param[out] Gk is the computed graph
  \param[out] g2gk_node is a mapping from the nodes of G to the nodes of Gk
  \param[out] gk2g_node is a mapping from the nodes of Gk to the nodes of G
 */
 void get_Gk(const DAG &G, int t, const DAG &PKG, const node_map<node> &g2pkg_node, const node_map<node> &pkg2g_node, const node_array<node> k, graph &Gk, node_map<node> &g2gk_node, node_map<node> &gk2g_node);

}

#endif
