/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

/*! \file architecture.cpp
\brief This files contains the implementation of all the methods of ARCHITECTURE class (describing an IA).
Each DAG of LOOP object should have an ISA (ARCHITECTURE) object attached to it.
*/
#include "architecture.h"
#include <iostream>
#include <fstream>
#include <LEDA/core/d_array.h>
#include "ddg_exceptions.h"

#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMDocumentType.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMNodeIterator.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/dom/DOMText.hpp>

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/XMLUni.hpp>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

XERCES_CPP_NAMESPACE_USE
        
namespace DDG {

// create an empty architecture
/*!
	\fn DDG::ARCHITECTURE::ARCHITECTURE()
	Create an empty ISA object. By default, the code semantic is set to NUAL.
 */
ARCHITECTURE::ARCHITECTURE(): inst_types(), reg_types()
{
cpt_regtypes=0; // a counter incremented for each new register type
cpt_instructions_types=0;
default_arch=false;
is_ual_semantic = false;
}


ARCHITECTURE::~ARCHITECTURE()
{

}

/*!
    \fn DDG::ARCHITECTURE::read_from_xml()
    Reads an ISA object from an XML file, see \ref ISA_xml. 
  \returns true if everything is ok
 */
bool ARCHITECTURE::read_from_xml(const char * xml_file)
{
	default_arch=false; // Fred, laisse moi ce flag
  bool result = false;
  
  try{
    XMLPlatformUtils::Initialize();
  } catch ( ... ){
    cerr << "DDG - Error while initializing XML : a default architecture is assumed" << endl;
    set_default_arch();
    return false;
  }
   this -> clear();
  xercesc::XercesDOMParser *config_parser = new XercesDOMParser;
  // Tags and attributes we are interested in parsing.
  XMLCh* TAG_Arch        = XMLString::transcode("arch");
  XMLCh* TAG_Instruction = XMLString::transcode("instruction");
  XMLCh* TAG_Write       = XMLString::transcode("write");
  XMLCh* TAG_Regtype     = XMLString::transcode("regtype");
  XMLCh* TAG_Register    = XMLString::transcode("register");
  XMLCh* TAG_Semantic    = XMLString::transcode("semantic");
  
  XMLCh* ATTR_Version   = XMLString::transcode("version");
  XMLCh* ATTR_Opcode	  = XMLString::transcode("opcode");
  XMLCh* ATTR_Latency   = XMLString::transcode("latency");
  XMLCh* ATTR_Regtype   = XMLString::transcode("regtype");
  XMLCh* ATTR_Delta_w   = XMLString::transcode("delta_w");
  XMLCh* ATTR_Delta_r   = XMLString::transcode("delta_r");
  XMLCh* ATTR_Type      = XMLString::transcode("type");
  XMLCh* ATTR_Number    = XMLString::transcode("number");
  XMLCh* ATTR_Name      = XMLString::transcode("name");
  
  XMLCh* TYPE_UAL       = XMLString::transcode("UAL");

  try{
    // Check if file exists and can be read
    struct stat filebuf;
    int status;
    status = stat(xml_file, &filebuf);
    
    if (status != 0){
      cerr << "DDG error: Cannot stat XML file" << endl;
      InvalidISADesc inv;
      throw (inv);
    }
    
    // @todo : use DTD
    config_parser->setValidationScheme( XercesDOMParser::Val_Auto);
    
    config_parser->parse(xml_file);
    
    DOMDocument* xml_doc = config_parser->getDocument();
    
    // Get the top-level element : "arch"
    DOMElement* root = xml_doc->getDocumentElement();
    if( !root ){
      cerr << "DDG error in architecture parsing (XML): Error with input XML file" << endl;
      InvalidISADesc inv;
      throw (inv);
    }
    
    int version  = XMLString::parseInt(root->getAttribute(ATTR_Version));
    
    // @todo Only support version 1
    if(version != 1){
      cerr << "DDG error in architecture parsing (XML): Wrong XML version " << endl;
      InvalidISADesc inv;
      throw (inv);
    }
   
    // First, get the register types
    DOMNodeList* regtypes = root->getElementsByTagName(TAG_Regtype);
    XMLSize_t node_count = regtypes->getLength();
    
    for( XMLSize_t x = 0; x < node_count; ++x ) {
      DOMNode* current_node = regtypes->item(x);
      if( current_node->getNodeType() && current_node->getNodeType() == DOMNode::ELEMENT_NODE ){
        // Found node which is an Element. Re-cast node as element
        DOMElement* current_element = dynamic_cast< xercesc::DOMElement* >( current_node );
        
        // Create the new regtype
        char *type = XMLString::transcode(current_element->getAttribute(ATTR_Type));
        int number = XMLString::parseInt(current_element->getAttribute(ATTR_Number));
        REGISTER_TYPES rt(std::string(type), number);
        XMLString::release(&type);
        rt.set_register_type_id(create_new_regtype_id());
        
        // Find register names
        DOMNodeList* registers = current_element->getElementsByTagName(TAG_Register);
        const  XMLSize_t subnode_count = registers->getLength();
        for(XMLSize_t y = 0; y < subnode_count; ++y){
          DOMNode* current_subnode = registers->item(y);
          if( current_subnode->getNodeType() && current_subnode->getNodeType() == DOMNode::ELEMENT_NODE ){
            DOMElement* current_subelement = dynamic_cast< xercesc::DOMElement* >( current_subnode );
            char *reg_name = XMLString::transcode(current_subelement->getAttribute(ATTR_Name));
            rt.add_register_name(std::string(reg_name));
            XMLString::release(&reg_name);
          }
        }
        
        // Add regtype to architecture
        new_register_type(rt);
      }
    }
   
    // Get instructions, checking they refer to existing regtypes
    DOMNodeList* instructions = root->getElementsByTagName(TAG_Instruction);
    node_count = instructions->getLength();
    
    for( XMLSize_t x = 0; x < node_count; ++x ) {
      DOMNode* current_node = instructions->item(x);
      if( current_node->getNodeType() && current_node->getNodeType() == DOMNode::ELEMENT_NODE ){
        DOMElement* current_element = dynamic_cast< xercesc::DOMElement* >( current_node );
        
        // Create a new instruction
        char *opcode = XMLString::transcode(current_element->getAttribute(ATTR_Opcode));
        int latency = XMLString::parseInt(current_element->getAttribute(ATTR_Latency));
        int delta_r = XMLString::parseInt(current_element->getAttribute(ATTR_Delta_r));
        INSTRUCTIONS_TYPES it(opcode, create_new_instruction_type_id(), latency);
        XMLString::release(&opcode);
        it.set_delta_r(delta_r);
        
        // Find what regtypes the instruction writes into
        DOMNodeList* writes = current_element->getElementsByTagName(TAG_Write);
        const  XMLSize_t subnode_count = writes->getLength();
        for(XMLSize_t y = 0; y < subnode_count; ++y){
          DOMNode* current_subnode = writes->item(y);
          if( current_subnode->getNodeType() && current_subnode->getNodeType() == DOMNode::ELEMENT_NODE ){
            DOMElement* current_subelement = dynamic_cast< xercesc::DOMElement* >( current_subnode );
            char *reg_type = XMLString::transcode(current_subelement->getAttribute(ATTR_Regtype));
            int delta_w = XMLString::parseInt(current_subelement->getAttribute(ATTR_Delta_w));
            int id = get_register_type_id(std::string(reg_type));
            XMLString::release(&reg_type);
            if(id == -1){
              cerr << "DDG error in architecture parsing (XML). Undefined register type" << endl;
              InvalidISADesc inv;
              throw (inv);
            }
            it.add_write_reg_id(id);
            it.set_delta_w(id, delta_w);
          }
        }
        // Add instruction to architecture
        new_instruction_type(it);
      }
    }

    // Get semantic : only one such tag is allowed
    DOMNodeList* semantic = root->getElementsByTagName(TAG_Semantic);
    node_count = semantic->getLength();

    if(node_count == 1){
      DOMNode* current_node = semantic->item(0);
      if( current_node->getNodeType() && current_node->getNodeType() == DOMNode::ELEMENT_NODE ){
        DOMElement* current_element = dynamic_cast< xercesc::DOMElement* >( current_node );
        if(XMLString::equals(current_element->getAttribute(ATTR_Type), TYPE_UAL)){
          set_ual_semantic(true);
        }
      }
    } /*else {
      cerr << "DDG  architecture parsing (XML): Code semantics was not specified, using NUAL" << endl;
    }*/

    result = check();
  // Fall back on default arch if there are errors
  }catch( ... ){
    cerr<<"DDG error  in architecture parsing : a default architecture description is assumed."<<endl;
		set_default_arch();
		result = false;
  }

  // Clean up
  XMLString::release(&TAG_Arch       );        
  XMLString::release(&TAG_Instruction);
  XMLString::release(&TAG_Write      );
  XMLString::release(&TAG_Regtype    );
  XMLString::release(&TAG_Register   );
  XMLString::release(&TAG_Semantic   );
  XMLString::release(&ATTR_Version   );
  XMLString::release(&ATTR_Opcode	   );
  XMLString::release(&ATTR_Latency   );
  XMLString::release(&ATTR_Regtype   );
  XMLString::release(&ATTR_Delta_w   );
  XMLString::release(&ATTR_Delta_r   );
  XMLString::release(&ATTR_Type      );
  XMLString::release(&ATTR_Number    );
  XMLString::release(&ATTR_Name      );
  XMLString::release(&TYPE_UAL       );
  XMLPlatformUtils::Terminate();
  return result; 
}


/*!
 * \fn DDG::ARCHITECTURE::read_from_arch()
 * \deprecated
 * Works ontly for a simple model of the architecture : a unique register type; implicitely, it is the regtype named "general".
 * \returns true if everything is ok
 * \ref ISA_desc
 * \throws InvalidISADesc When the ISA description file is corrupted, and object of class DDG::InvalidISADesc is thrown.
 */
bool ARCHITECTURE::read_from_arch(const char * arch_desc) {
  LEDA::list<INSTRUCTIONS_TYPES>::item it;
  int i, j, k, l, id_reg=create_new_regtype_id();
  bool is_value;
  char c;
  std::string str;
  char cc[MAX_STRING];
  
  ifstream desc(arch_desc);
  default_arch=false;
  clear();
  
  // create a unique register type named general
  REGISTER_TYPES rt("general");
  rt.set_register_type_id(id_reg); // set the register type id
  new_register_type(rt); // add the register type to the architecture
  
  // read instruction types from the arch file
  try{
    if(desc) {
      
      desc.getline(cc, MAX_STRING, '\n'); // get the header section
      str=std::string(cc);
      i=str.find("DDG.ARCH.DESCRIPTION"); // looks for the header section specifier
      if ((i<0)||i>str.size()) {
        cerr <<"DDG bad header section : "<<arch_desc<<". Seems not a valid ISA description.\n";
        desc.close();
			InvalidISADesc inv;
			throw (inv);
		}
		while (desc.good()){
			INSTRUCTIONS_TYPES it;
			desc>>str>>i>>is_value>>j>>k>>l; // read the str_opcode, the int opcode write_into-reg?, latency, reading delay, writing delay;
			it=INSTRUCTIONS_TYPES(str, i, j);
			it.set_delta_r(k);

			if(is_value){ 
				// if the instruction writes into the unique register type
			 	it.add_write_reg_id(id_reg); 
			 	it.set_delta_w(id_reg, l);
			}
			new_instruction_type(it);// insert it in the architecture 
			
			cpt_instructions_types=leda::max(cpt_instructions_types, i)+1; // update the counter of instructions type
		}
	}
	else{
		cerr<<"DDG IO Error: Could not read the architecture description file: "<<arch_desc<<endl;
		InvalidISADesc inv;
		throw (inv);
		
	}
	}
	catch(InvalidISADesc)
	{
		cerr<<"DDG : a default architecture description is assumed."<<endl;
		// instructions types by default {add_op=0, sub_op=1, not_op=2, mul_op=3, div_op=4, ld_op=5, st_op=6, copy_op=7, nop_op=8}; 
		// nops and store instructions do not write into registers
		
		set_default_arch();
		return false;
	}
	
	// make a final check to see if the parsed arch is OK
	return (check());
}


/*!
    \fn DDG::ARCHITECTURE::clear()
    Makes empty the ISA object.
 */
void ARCHITECTURE::clear()
{ 
inst_types.clear();
reg_types.clear();
default_arch=false;
}

/*!
    \fn DDG::ARCHITECTURE::write_to_arch(char * filename)
    \deprecated
    Dump the architecture to an arch file (deprecated format).
 */
void ARCHITECTURE::write_to_arch(const char * filename) const
{
  FILE *fp;

  try{
    fp = fopen(filename, "w");
    if(fp == NULL){
      cerr << "Could not open file" << endl;
      BadIODDG bioddg;
      throw(bioddg);
    }
    fprintf(fp, "DDG.ARCH.DESCRIPTION 1.1\n");
    INSTRUCTIONS_TYPES it;
    forall(it, inst_types){
      int num = it.get_list_reg_id().length();
      int delta_w = 0;
      int write_into = 0;
      if(num == 1){
        delta_w = it.delta_w(it.get_list_reg_id().pop());
        write_into = 1;
      }
      if(num > 1){
        cerr << "Error : Arch file format is not compatible with more than one registr type\n";
        NonUniqueRegisterType nurt;
        throw(nurt);
      }
      fprintf(fp, "%s\t%d\t%d\t%d\t%d\t%d\n", it.opcode().c_str(), it.opcode_id(), write_into, it.latency(), it.delta_r(), delta_w); 
    }
   
    fclose(fp);

  }catch( ... ){
    cerr << "Error while writing file : " << filename << endl;
  }
} 


/*!
    \fn DDG::ARCHITECTURE::write_to_xml(char*)
    Dump the ISA object into an an XML file, see \ref ISA_xml
 */
void ARCHITECTURE::write_to_xml(const char *xml_file) const
{
 
 ofstream fp;
 LEDA::list<std::string> list_reg;
 LEDA::list<int> list_reg_id;
 try{
  fp.open(xml_file);
  
  fp << "<arch version='1'>\n";
 
  REGISTER_TYPES reg;
  forall(reg, reg_types){
    fp << "<regtype type='" << reg.get_register_typename() << "' number='" << reg.get_nb_registers() << "'>\n";
    std::string name;
    list_reg=reg.get_list_registers();
    forall(name, list_reg){
      fp << "<register name='" << name << "'/>\n";
    }
    fp << "</regtype>\n";
  }

  INSTRUCTIONS_TYPES it;
  forall(it, inst_types){
    fp << "<instruction opcode='" << it.opcode() << "' latency='" << it.latency() << "' delta_r='" << it.delta_r() << "'>\n";
    int id;
    list_reg_id=it.get_list_reg_id();
    forall(id, list_reg_id){
      reg = get_register_type_by_id(id);
      fp << "<write regtype='" << reg.get_register_typename() << "' delta_w='" << it.delta_w(id) << "'/>\n";
    }
    fp << "</instruction>\n";
  }
 
  if (is_ual_semantic){
    fp << "<semantic type='UAL'/>\n";
  }
  fp << "</arch>\n";
  
  fp.close();
 }catch( ... ){
  cerr << "Error while writing file : " << xml_file << endl;
 }
}


/*!
    \fn DDG::ARCHITECTURE::check()
    Simple check method. 
\returns true if everything seems OK. False otherwise.
 */
bool ARCHITECTURE::check() const
{
 	bool res_check=true;
	
    	INSTRUCTIONS_TYPES instr;
	REGISTER_TYPES reg;
	LEDA::list<int> inst_lregtypeid;
	int regtype, i;
	
	// first check if there are any instructions types or registers types
	if (inst_types.empty()&&reg_types.empty()){
		cerr<<"DDG warning : Architecture check finds an empty architecture.\n";
		return false;
	}
	//first check every register type individually
	forall(reg,reg_types){
		res_check=reg.check();
		if (res_check==false) return false;
		// here we build the list of all regtypes id
	}
	
	//then check every instruction type individually
	forall(instr, inst_types){
		res_check=instr.check();
		if (res_check==false) return false;
		// check if the regitype id referenced by the instruction type is OK
		inst_lregtypeid=instr.get_list_regtypes_id();
		forall(regtype, inst_lregtypeid){
			if (!exist_register_type_id(regtype)){
			cerr<<"DDG warning : Architecture check for the opcode "<<instr.opcode()<<" : " << " The id of the regtype seems to not exist in the list of regtypes.\n";
			return false;
			}
		}
	}
	
	//if I reach this point, then the arch description is fine
	return true;
}



/*!
    \fn DDG::ARCHITECTURE::get_list_regtype_id()
    \returns the list of the id of the registers types
 */
LEDA::list<int> ARCHITECTURE::get_list_regtypes_id() const
{
    	LEDA::list<int> lregtypeid;
    	REGISTER_TYPES reg;
    	forall(reg,reg_types){
		lregtypeid.push(reg.get_register_type_id());
	}
	return lregtypeid;
}

/*!
    \fn DDG::ARCHITECTURE::get_list_instructions_types_id
    \returns the list of the id of the instructions types
 */
LEDA::list<int> ARCHITECTURE::get_list_instructions_types_id() const
{
	LEDA::list<int> linstid;
	INSTRUCTIONS_TYPES it;
	
	forall(it,inst_types){
		linstid.push(it.opcode_id());
	}
	
	return linstid;
}


/*!
    \fn DDG::ARCHITECTURE::get_register_type_id(std::string regtype)
    \returns the id of a register type given by its string name. Returns -1 if it does not exist.
 */
int ARCHITECTURE::get_register_type_id(std::string regtype_name) const
{
    REGISTER_TYPES rt;
    forall(rt, reg_types){
    	if (rt.get_register_typename() == regtype_name) return rt.get_register_type_id();
    }
    return (-1);
}


/*!
    \fn DDG::ARCHITECTURE::get_register_typename(int regtype_id)
    \returns the string name of a register type given by its id. Returns empty string if it does not exist.
 */
std::string ARCHITECTURE::get_register_typename(int regtype_id) const
{
    REGISTER_TYPES rt;
    forall(rt, reg_types){
    	if (rt.get_register_type_id() == regtype_id) return rt.get_register_typename();
    }
    return std::string("");
}


/*!
    \fn DDG::ARCHITECTURE::exist_register_type_id(int registertype_id)
    \returns true if the given register type id exists in the architecture.
 */
bool ARCHITECTURE::exist_register_type_id(int registertype_id) const
{
  return reg_types.defined(registertype_id);
    
}

/*!
    \fn DDG::ARCHITECTURE::exist_instruction_type_id(int id)
    \returns true if the given instruction type id exists in the architecture.
 */
bool ARCHITECTURE::exist_instruction_type_id(int id) const
{
  return this->inst_types.defined(id);  
}



/*!
    \fn DDG::ARCHITECTURE::exist_register_typename(std::string reg_typename)
    \returns true if the given register typename exists in the architecture.
 */
bool ARCHITECTURE::exist_register_typename(std::string reg_typename) const
{
    REGISTER_TYPES rt;
    forall(rt, reg_types){
    	if (rt.get_register_typename() == reg_typename) return true;
    }
    return (false);
}

/*!
    \fn DDG::ARCHITECTURE::exist_register_typename(std::string reg_typename)
    \returns true if the given opcode exists in the architecture
 */
bool ARCHITECTURE::exist_instruction_type_opcode(std::string stropcode) const
{
    INSTRUCTIONS_TYPES it;
    forall(it, inst_types){
    	if (it.opcode() == stropcode) return true;
    }
    return (false);
}


/*!
    \fn DDG::ARCHITECTURE::get_new_regtype_id()
    \returns a unique id to assign to a new register type. This function must be used to get a unique register type id to add inside the architecture.

 */
int ARCHITECTURE::create_new_regtype_id()
{
	REGISTER_TYPES rt;
	cpt_regtypes++;
	rt.set_register_type_id(cpt_regtypes);
	reg_types[cpt_regtypes]=rt;
    	return cpt_regtypes;
}


/*!
    \fn DDG::ARCHITECTURE::get_new_instruction_type_id()
   \returns a unique id to assign to an instruction type. This function must be used to get a unique instruction type id to add inside the architecture.
 */
int ARCHITECTURE::create_new_instruction_type_id()
{
	INSTRUCTIONS_TYPES it;
	cpt_instructions_types++;
	it.set_opcode_id(cpt_instructions_types);
	inst_types[cpt_instructions_types]=it;
    	return cpt_instructions_types;
}


/*!
    \fn DDG::ARCHITECTURE::new_register_type(REGISTER_TYPES rt)
    Add a new register type in the architetcure. 
    \remarks Be carefull, a unique id for the register type should be set before in the rt object. Use the function DDG::ARCHITECTURE::get_new_regtype_id() to get a unique id.
 */
void ARCHITECTURE::new_register_type(REGISTER_TYPES rt)
{
	default_arch=false;
	reg_types[rt.get_register_type_id()]=rt;
}




/*!
    \fn DDG::ARCHITECTURE::new_instruction_type(INSTRUCTIONS_TYPES it)
     Add a new instruction type in the architetcure. 
     \remarks Be carefull, a unique id for the instruction type should be set before in the it object. Use the function DDG::ARCHITECTURE::get_new_regtype_id() to get a unique id.
 */
void ARCHITECTURE::new_instruction_type(INSTRUCTIONS_TYPES it)
{
    default_arch=false;
    inst_types[it.opcode_id()]=it;
}

/*!
    \fn DDG::ARCHITECTURE::set_default_arch()
    Set a default simple architecture (built in by DDG). 
    By default, we have a unique register type (general), and 9 operations : add, sub, not, mult, div, ld, st, copy, nop.
 */
void ARCHITECTURE::set_default_arch()
{
	INSTRUCTIONS_TYPES it;
	int id_regtype= create_new_regtype_id(); // create a new register type id
	int id_inst;
	clear();
	REGISTER_TYPES rt("general");
	rt.set_register_type_id(id_regtype); // set the register type id
	new_register_type(rt); // add the register type to the architecture
	
	it= INSTRUCTIONS_TYPES("add_op",0); // create an add op.with its id
	it.add_write_reg_id(id_regtype, 0); // insert the register type in the instruction type
	new_instruction_type(it); // insert the instruction to the architecture
	
	
	it= INSTRUCTIONS_TYPES("sub_op", 1); // create an sub op.with its id
	it.add_write_reg_id(id_regtype, 0); // insert the register type in the instruction type
	new_instruction_type(it); // insert the instruction to the architecture
	
	
	it= INSTRUCTIONS_TYPES("not_op", 2); // create a not op.with its id
	it.add_write_reg_id(id_regtype, 0); // insert the register type in the instruction type
	new_instruction_type(it); // insert the instruction to the architecture
	
	
	it= INSTRUCTIONS_TYPES("mul_op", 3); // create a mult op.with its id
	it.add_write_reg_id(id_regtype, 0); // insert the register type in the instruction type
	new_instruction_type(it); // insert the instruction to the architecture
	
	
	it= INSTRUCTIONS_TYPES("div_op", 4); // create a div op.with its id
	it.add_write_reg_id(id_regtype, 0); // insert the register type in the instruction type
	new_instruction_type(it); // insert the instruction to the architecture
	
	it= INSTRUCTIONS_TYPES("ld_op", 5); // create a ld op.with its id
	it.add_write_reg_id(id_regtype, 0); // insert the register type in the instruction type
	new_instruction_type(it); // insert the instruction to the architecture
	
	
	it= INSTRUCTIONS_TYPES("st_op", 6); // create a st  op.with its id
	new_instruction_type(it); // insert the instruction to the architecture
	
	it= INSTRUCTIONS_TYPES("copy_op", 7); // create a copy  op.with its id
	it.add_write_reg_id(id_regtype, 0); // insert the register type in the instruction type
	new_instruction_type(it); // insert the instruction to the architecture
	
	it= INSTRUCTIONS_TYPES("nop_op", 8); // create a nop.with its id
	new_instruction_type(it); // insert the instruction to the architecture
			
	// set the global counter of id
	cpt_instructions_types=9;
	default_arch=true;
}


/*!
    \fn DDG::ARCHITECTURE::number_of_registers_types()
    \returns the number of register types present in the ISA.
 */
int ARCHITECTURE::number_of_registers_types() const
{
    return reg_types.size();
}


/*!
    \fn DDG::ARCHITECTURE::get_unique_register_type_id()
    \pre A unique register type should exist in the architecture
    \throws NonUniqueRegisterType
 */
int ARCHITECTURE::get_the_unique_register_type_id() const
{
	REGISTER_TYPES rt;
	if (number_of_registers_types()!=1){
		throw NonUniqueRegisterType();
	}
	forall(rt, reg_types) {
		return (rt.get_register_type_id());
	}
}


/*!
    \fn DDG::ARCHITECTURE::get_instruction_type_by_opcode()
    \returns An INSTRUCTIONS_TYPES object with opcode_id equal to -1 if not found
 */
INSTRUCTIONS_TYPES ARCHITECTURE::get_instruction_type_by_opcode(std::string stropcode) const
{
	INSTRUCTIONS_TYPES it;
	forall(it, inst_types){
		if(it.opcode()== stropcode) return it;
	}
	it.set_opcode_id(-1);
	return it;
}

/*!
    \fn DDG::ARCHITECTURE::get_list_register_typers()
    \returns the list of register types present in the ISA.
 */
LEDA::list<REGISTER_TYPES> DDG::ARCHITECTURE::get_list_register_types() const
{
    	LEDA::list<REGISTER_TYPES> lregtype;
    	REGISTER_TYPES reg;
    	forall(reg,reg_types){
        lregtype.push(reg);
      }
      return lregtype;
}


}





/*!
    \fn DDG::ARCHITECTURE::pick_an_instruction_type_id()
 */
int DDG::ARCHITECTURE::pick_an_instruction_type_id() const
{
    INSTRUCTIONS_TYPES it;
    forall(it, inst_types){
    	return it.opcode_id();
    }
    
    throw BadInstructionType();
}
