/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

#ifndef __architecture_H
#define __architecture_H

/*! \file architecture.h
\brief The header file of the C++ ARCHITECTURE class. 
This file containts all function prototypes for building an ISA (architecture) object: register types, instructions types, latencies, read and written registers for each instruction type, code semantics (UAL/NUAL). 
An ISA object can be constructed by XML import, or by C++ interface using the different class methods.
*/


#endif
#ifndef DDGARCHITECTURE_H
#define DDGARCHITECTURE_H

#include <LEDA/core/list.h>
#include <LEDA/core/d_array.h>
#include "ddg_exceptions.h"

#include "instructions_types.h"
/** \internal
\def MAX_STRING 
The maximal size of a char* is statically limited to 1000.
 */
#ifndef MAX_STRING
#define MAX_STRING 1000
#endif
namespace DDG {

/**
	@author Touati Sid <Sid-nospam.Touati-nospam@inria.fr>
*/
class ARCHITECTURE{
protected:
int cpt_regtypes; // a counter incremented for each new register type
int cpt_instructions_types; // a counter incremented for each new
bool default_arch;
bool is_ual_semantic;
private:	
LEDA::d_array<int,REGISTER_TYPES> reg_types;
LEDA::d_array<int,INSTRUCTIONS_TYPES> inst_types;
public:
    ARCHITECTURE();
    ~ARCHITECTURE();
    bool read_from_xml(const char *);
    bool read_from_arch(const char *);
    void clear();
    bool check() const;
    void write_to_xml(const char*) const;
    LEDA::list<int> get_list_regtypes_id() const;
    LEDA::list<int> get_list_instructions_types_id() const;
    LEDA::d_array<int,REGISTER_TYPES>  get_dictionary_regtypes() const {return reg_types;}
    LEDA::d_array<int,INSTRUCTIONS_TYPES> get_dictionary_instructions_types() const {return inst_types;}
  
    
    int get_register_type_id(std::string regtype_name) const;
    std::string get_register_typename(int regtype_id) const;
    bool exist_register_typename(std::string reg_typename) const;
    bool exist_register_type_id(int registertype_id) const;
    bool exist_instruction_type_id(int id) const;
    bool exist_instruction_type_opcode(std::string opcode) const;
    
    int create_new_instruction_type_id();
    int create_new_regtype_id();
    void set_default_arch();
    void new_register_type(REGISTER_TYPES rt);
    void new_instruction_type(INSTRUCTIONS_TYPES it);

    /*!
        \fn DDG::ARCHITECTURE::is_default_arch()
        Returns true if the architecture is the default one
     */
    bool is_default_arch()
    {
        return default_arch;
    }

    /*!
        \fn DDG::ARCHITECTURE::get_ual_semantic
        Returns true if semantic is UAL
     */
    bool get_ual_semantic() const{
      return is_ual_semantic;
    }
    
    /*!
        \fn DDG::ARCHITECTURE::set_ual_semantic(bool)
        Set to true if semantic is UAL
     */
    void set_ual_semantic(bool ual){
      is_ual_semantic = ual;
    }
    
    /*!
        \fn DDG::ARCHITECTURE::get_register_type_by_id(inr id)
        \pre the register type id exists
        \throw BadRegType
     */
    REGISTER_TYPES get_register_type_by_id(int id) const
    {
        if (exist_register_type_id(id)) return reg_types[id];
        else {
        	throw BadRegType();
        }
    }

    /*!
        \fn DDG::ARCHITECTURE::get_instruction_type_by_id(int id)
        \pre the instruction type id exists
        \throw BadInstructionType
     */
    INSTRUCTIONS_TYPES get_instruction_type_by_id(int id) const
    {
        if (exist_instruction_type_id(id)) return inst_types[id];
        else throw BadInstructionType(id);
    }
    int number_of_registers_types() const;
    int number_of_instructions_types() const;
    int get_the_unique_register_type_id() const;
    INSTRUCTIONS_TYPES get_instruction_type_by_opcode(std::string opcode) const;
    void write_to_arch(const char * filename) const;
    LEDA::list<REGISTER_TYPES> get_list_register_types() const;
    /*!
	\fn LEDA::list<REGISTER_TYPES> DDG::ARCHITETCURE::T()
         Same as get_list_register_types(). Returns the list of register types.	    
     */
    LEDA::list<REGISTER_TYPES> T() const {return get_list_register_types();}
    int pick_an_instruction_type_id() const;

    /*!
        \fn DDG::ARCHITECTURE::get_instruction_type_id(std::string)
    	Returns the id of the instruction type given by its opcode string.
     */
    int get_instruction_type_id(std::string stropcode)const
    {
	    return (this->get_instruction_type_by_opcode(stropcode)).opcode_id();
    }
   
    
};

}

#endif
