/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

/*! \file dag_example.cpp
\brief A program example showing how to use the DAG class.
*/

#include "DDG.h" // the header file of the DDG library. 
		// To be included by the user.
#include "LEDA/graph/graph_misc.h" // Is_Acyclic 
#include <iostream>
#include <cstdlib>


using namespace std;
using namespace DDG; // the namespace of the DDG library.



int main(int argc, char *argv[])
{
	//int are always usefull
	int         i,j,cpt ;
	int c; // for getopt
	// some strings for file names
	char        *ddg_filename=NULL, *arch_filename=NULL;
	char        str[MAX_STRING];
	REGISTER_TYPES rt;
	LEDA::list<REGISTER_TYPES> T;
	
	// some objects for graphs
	//declare a DDG of a loop (cyclic data dependence graph)
	DDG::LOOP loop; 
	ARCHITECTURE ISA_example;
	DDG::DAG dag, dag_copy;  // declare a DDG of a basic block (directed acyclic graph : DAG)
	
	LEDA::node n,u;
	LEDA::edge e;
	LEDA::set<node> VR;
	LEDA::set<node> MA; // for a maximal antichain computation
	LEDA::node_map<int> map; // when duplicating a graph, 
				// we need to save node correspondance between original and new graph
	LEDA::node_array<node> killer; // for register saturation examples
	array< list<node> > ALN ; // array of sorted chains

	DDG::INSTRUCTIONS_TYPES it; // information attached to an opcode
	
	LEDA::list<edge> el; 
	LEDA::list<node> nl; 
	LEDA::set<node> parents, children; // set of parents and children of a node

	while ((c = getopt (argc, argv, "i:a:")) != -1){
		switch(c){
			case 'a': arch_filename=optarg;
				break;
			case 'i': ddg_filename =optarg;
				break;
			case '?': 
			 	cout<<"usage:"<< argv[0] <<" -i ddg_filename.xml [-a isa_desc.xml] "<<endl;
			 	return EXIT_FAILURE;
		}
		
	}
	
	//------- the user can define its own ISA ----------------
	if (arch_filename!=NULL)  { 
		ISA_example.read_from_xml(arch_filename);
		if(ISA_example.check()){
			loop=LOOP(ISA_example);	 //build an empty DDG loop with a user defined architcural (ISA) description
			dag=DAG(ISA_example);	//build an empty DAG with a user defined architcural (ISA) description
		} else return EXIT_FAILURE;
	}
	 
	 //---------------- The user can read a DDG from an xml format defined by the DDG library 
	
	if (ddg_filename != NULL){
	
		i=loop.read_from_xml(ddg_filename); // reads the DDG loop from a n xml file
		
 		if(i==-1){
 			return EXIT_FAILURE;
 		}
		
		i=dag.read_from_xml(ddg_filename); // reads the DDG loop from a n xml file
		
		if(i==-1){
			return EXIT_FAILURE;
		}
		
	}
	else{
	    cout<<"usage:"<< argv[0] <<" -i ddg_filename.xml [-a isa_desc.xml] "<<endl;
	    return EXIT_FAILURE;
	}

	if(! Is_Acyclic(dag, el)) {
		cerr<<"Input Error: Non acyclic graph.\n";
		cerr<<"There are "<< el.size() <<" illegal arcs:\n";
		forall(e, el){
			cerr<<dag.id(dag.source(e))<<" -> "<< dag.id(dag.target(e))<<endl;
		}
		return EXIT_FAILURE;
	}

	
	 // The user can modify the DDG by many other methods : 
	// new_node, new_edge, clear, del_node, del_edge, etc.
	 // The user can access any information about nodes, edges,
	 // apply well known algorithms using the LEDA library
	 // See the LEDA manual : http://www.algorithmic-solutions.info/leda_manual
	 // See the DDG user manual inside this distribution
  	

	cout<< "DAG example program: " <<ddg_filename<<endl;
	cout<< "-----------------------------" <<endl;
	T=dag.T();
	
	forall(rt,T){
			
		cout<<"Register Saturation of the DAG (loop body) for register type "<< rt.get_register_typename() <<" is " <<REGISTER_SATURATION (dag, rt.get_register_type_id(), nl)<<endl;
		cout<<"Saturating values of type "<<rt.get_register_typename()<<endl;
		forall(n, nl){
		it= dag.instruction_type(n);
		cout<< dag.id(n)<< " with opcode " << it.opcode()<< " ("<<dag.latency(n)<<","<<dag.delta_r(n)<<","<<dag.delta_w(n, rt.get_register_type_id())<<")"<<endl;
		}
	}
	// -------------------- The critical path is the longest path of the DAG
	cout<<endl<<"Critical Path ="<<dag.CriticalPath(el)<<endl;

	cout<<"Edges belonging to the critical path"<<endl;
	forall(e, el){
		n=dag.source(e);
		u=dag.target(e);
		it= dag.instruction_type(n);
		cout<< dag.id(n)<< " -> " << dag.id(u) <<endl;
	}

	//----------------------- We can access the DAG in many ways
	forall_nodes(u, dag){
		parents=dag.get_parents(u);
		cout<<"The parents of the node number "<<dag.id(u)<<" are : ";
		forall (n, parents){
			cout<<dag.id(n)<<" ";
		}
		cout<<endl;
	}
	forall_nodes(u, dag){
		children=dag.get_children(u);
		cout<<"The children of the node number "<<dag.id(u)<<" are : ";
		forall (n, children){
			cout<<dag.id(n)<<" ";
		}
		cout<<endl;
	}
	forall_nodes(u, dag){
		parents=dag.get_up(u);
		cout<<"The ascendants of the node number "<<dag.id(u)<<" are : ";
		forall (n, parents){
			cout<<dag.id(n)<<" ";
		}
		cout<<endl;
	}
	forall_nodes(u, dag){
		children=dag.get_down(u);
		cout<<"The descendants of the node number "<<dag.id(u)<<" are : ";
		forall (n, children){
			cout<<dag.id(n)<<" ";
		}
		cout<<endl;
	}
	
	// ------------------ The user can visualize the DAG ------------------------
	snprintf(str, MAX_STRING, "%s_dag.vcg",ddg_filename); // sets the .vcg filename for the DAG
	dag.write_to_vcg(str); //outputs the DAG to a vcg file to be visualized by the xvcg tool
	dag_copy.copy(dag); // create an exact copy.
	snprintf(str, MAX_STRING, "%s_dag_copy.vcg",ddg_filename); 
	dag_copy.write_to_vcg(str); //outputs the copy DAG to a vcg file to be visualized by the xvcg tool
	
	cout<<endl<<str<< " can be visualized using xvcg."<<endl;

	/*--------------- The user can extract a copy of a loop body -------------- */
	
	map.init(loop);/* initializes a mapping function from loop nodes to dag nodes.
	Forall u a node in the loop, map(u) stores its corresponding node in the exctracted DAG (loop body) */
	dag.clear();
	forall_nodes(u,loop){
		map[u]=dag.new_node((loop.instruction_type(u)).opcode_id());
	}
	forall_edges(e,loop){
		if(loop.lambda(e)==0) { // It is an edge belonging to the loop body ==> insert it into the DAG
			dag.new_edge(map[loop.source(e)], map[loop.target(e)],loop[e]);
		}
	}

	// --- Dilworth decomposition -----------
	i=MAXIMAL_ANTI_CHAIN(dag, MA);
	cout<<"Maximal antichain in the DAG : "<<i<<endl;
	cout<<"Nodes belonging to a maximal antichain "<<endl;
	forall(n, MA){
		cout<< dag.id(n)<< " with opcode " << it.opcode()<<endl;
	}

	//---The user can extract the body of a loop unrolled many times
	// (recurrent data dependence edges are preserved) ----------
	
	cout<<"Unrolling the loop "<<ddg_filename<<" four times and extract its body"<<endl;
	unroll (loop, dag, 4);
	snprintf(str, MAX_STRING,"%s_dag_unroll_4.vcg",ddg_filename); // sets the .vcg filename for the ddg loop
	dag.write_to_vcg(str);
	// The user can access any kind of information 
	forall(rt,T){
		cout <<"The set of values of type "<< rt.get_register_typename() <<" of the body  is :"<<endl;
		VR=dag.get_V_R(rt.get_register_type_id());
		forall(u, VR){
			it=dag.instruction_type(u);
			cout<< dag.id(u)<< " with opcode " << it.opcode()<<endl;
	
		}
	}
	cout<<endl<< "The set of sources of the body are :"<<endl;
	VR=dag.get_Sources();
	forall(u, VR){
		it=dag.instruction_type(u);
		cout<< dag.id(u)<< " with opcode " << it.opcode()<<endl;
	}
	cout<<endl<< "The set of sinks of the body are :"<<endl;
	VR=dag.get_Targets();
	forall(u, VR){
		it=dag.instruction_type(u);
		cout<< dag.id(u)<< " with opcode " << it.opcode()<<endl;
	}
	
	// Computes the critical path of the DAG and puts the list of its nodes in nl
	cout<<endl<<"Critical Path of the unrolled body ="<<dag.CriticalPath(nl)<<endl;
	cout<<"Nodes belonging the critical path of the unrolled body "<<endl;
	forall(n, nl){
		it= dag.instruction_type(n);
		cout<< dag.id(n)<< " with opcode " << it.opcode()<<endl;
	}

	// ------------- Hiding and restoring nodes -------------
	// look for two arbitrary values of any register type
	forall(rt, T){
		if (dag.get_V_R(rt.get_register_type_id()).size()>=2){
			dag.hide_node(dag.get_V_R(rt.get_register_type_id()).choose()); // just to hide one node writing into a register
						// note that all adjacent edges are hidden too
			dag.hide_node(dag.get_V_R(rt.get_register_type_id()).choose()); // hide another one
			break;
		}
	}
	rt=T.front();
	cout<<"Register Saturation of the DAG (after hiding two values of type "<< rt.get_register_typename() <<") is : "<<(j=REGISTER_SATURATION (dag, rt.get_register_type_id(), nl, killer))<<endl;
	cout<<"Saturating values of type "<< rt.get_register_typename() <<endl;
	forall(n, nl){
		it= dag.instruction_type(n);
		cout<< dag.id(n)<< " with opcode " << it.opcode()<<endl;
	}
	
	dag.restore_all_nodes(); 
	dag.restore_all_edges();
	dag.build_internal_structures(); // restore the internal state of the DAG

	cout<<"Register Saturation of the DAG (after restoring the two values of type"<< rt.get_register_typename()<<") is : "<<REGISTER_SATURATION (dag, rt.get_register_type_id(), nl, killer)<<endl;
	cout<<"Saturating values of type "<< rt.get_register_typename()<<endl;
	forall(n, nl){
		it= dag.instruction_type(n);
		cout<< dag.id(n)<< " with opcode " << it.opcode()<<endl;
	}

	//--------------------- Lattices and Orders --------------------------/
	cout<<"Lattices and Order Notations:\n";
	forall_nodes(u, dag){
		forall_nodes(n, dag){
			if(lt(u,n)){
				cout<<dag.id(u)<< " < " << dag.id(n) << endl;
			}
			if(le(u,n)){
				cout<<dag.id(u)<< " <= " << dag.id(n)<< endl;
			}
			if(gt(u,n)){
				cout<<dag.id(u)<< " > " << dag.id(n)<< endl;
			}
			if(ge(u,n)){
				cout<<dag.id(u)<< " >= " << dag.id(n)<< endl;
			}
			if(parallel(u,n)){
				cout<<dag.id(u)<< " || " << dag.id(n)<< endl;
			}
			if(comparable(u,n)){
				cout<<dag.id(u)<< " ~ " << dag.id(n)<< endl;
			}
		}
	}


	cout<<"Minimal Chain Decomposition (Dilworth)"<<endl;
	cout<<"--------------------------------------"<<endl;
	j=MINIMAL_CHAIN(dag, ALN);

	cout<<"There are "<< j <<" chains"<<endl;
	cpt=0;
	for(i=0;i< j ;i++){
		cout<<"Chain["<<i<<"]: ";
		forall(u,ALN[i]){
			cout<<dag.id(u);
			cout<<" ";
			cpt++;
		}
		cout<<endl;
	}
	cout<<"Total "<< cpt <<" nodes"<<" in this distribution."<<endl;
	//--------------------- Many other software opportunities exist. See the reference manual.

	return EXIT_SUCCESS;
}
