/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

/*! \file info_edge.cpp
\brief Methods implementations of the INFO_EDGE class.
The INFO_EDGE class contains all attributes attached to a DDG edge. 
*/
#include "info_edge.h"
#include "ddg_exceptions.h"

using namespace std;
namespace DDG{

/*!
	\fn bool operator==(const INFO_EDGE u, const INFO_EDGE v)
	Returns true iff \f$ \delta(u)= \delta(v) \f$ and \f$ \lambda(u)= \lambda(v) \f$
 */
bool operator==(const INFO_EDGE u, const INFO_EDGE v) {
  return ((u.delta()==v.delta())&&(u.lambda()==v.lambda()));
}

INFO_EDGE::INFO_EDGE() 
{duration=0;
	val_duration=0;
	kind=serial;
	distance=0;
	regtype_id=-1;
}

INFO_EDGE::INFO_EDGE(edge_type et)
{
	duration=0;
	val_duration=0;
	kind=et;
	regtype_id=-1;
}

INFO_EDGE::INFO_EDGE(int dur, int dis, edge_type et)
{
	duration=dur;
	val_duration=(double)dur;
	distance=dis;
	kind=et;
	regtype_id=-1;
}

INFO_EDGE::INFO_EDGE(int dur, edge_type et)
{
	duration=dur;
	val_duration=(double)dur;
	kind=et;
	distance=0;
	regtype_id=-1;
}


INFO_EDGE::INFO_EDGE(const INFO_EDGE &ie)
{
	distance=ie.distance;
	duration=ie.duration;
	val_duration=ie.val_duration;
	kind=ie.kind;
	regtype_id=ie.regtype_id;
}


/**
 * Converts a string (that represents a dependance type) to an edge_type
 * @param et 
 * @return The dependance type from its string description
 \throw 
 */
std::string  edge_type_to_string(const edge_type et) {
	if(et==flowdep_reg) return std::string("flowdep_reg");
	else if(et==serial) return std::string("serial");
	else if(et==antidep_reg) return std::string("antidep_reg");
	else if(et==outputdep_reg) return std::string("outputdep_reg");
	else if(et==inputdep_reg) return std::string("inputdep_reg");
	else if(et==flowdep_mem) return std::string("flowdep_mem");
	else if(et==antidep_mem) return std::string("antidep_mem");
	else if(et==outputdep_mem) return std::string("outputdep_mem");
	else if(et==inputdep_mem) return std::string("inputdep_mem");
	
	else if(et==spilldep_mem) return std::string("spilldep_mem");
	else if(et==other_mem) return std::string("other_mem");
	else if(et==killerdep) return std::string("killerdep");
	else if(et==reusedep) return std::string("reusedep");
	else throw IllegalDependance();
}


/**
 * Converts a string (that represents a dependance type) to an edge_type
 * @param et 
 * @return The dependance type from its string description
 \throw IllegalDependance
 */
edge_type string_to_edge_type(const std::string et) {
	if(et==std::string("flowdep_reg")) return flowdep_reg;
	else if(et==std::string("serial")) return serial;
	else if(et==std::string("antidep_reg")) return antidep_reg;
	else if(et==std::string("outputdep_reg")) return outputdep_reg;
	else if(et==std::string("inputdep_reg")) return inputdep_reg;
	else if(et==std::string("antidep_mem")) return antidep_mem;
	else if(et==std::string("outputdep_mem")) return outputdep_mem;
	else if(et==std::string("inputdep_mem")) return inputdep_mem;
	else if(et==std::string("flowdep_mem")) return flowdep_mem;
	
	else if(et==std::string("spilldep_mem")) return spilldep_mem;
	else if(et==std::string("other_mem")) return other_mem;
	else if(et==std::string("killerdep")) return killerdep;
	else if(et==std::string("reusedep")) return reusedep;
	else throw IllegalDependance();
	
}
ostream& operator<<(ostream &os,  INFO_EDGE x)  /* output stream */
{
	os<<x.delta()<<' '<< x.lambda()<<' '<< edge_type_to_string(x.etype());
	return os;
}
  
istream& operator>>(istream &is, INFO_EDGE & x)  /* input stream */
{
	std::string et;
	int i,j;
	
	is >> i >> j >> et; // read delta  lambda edge_type
	x.set_sigma(i);
	x.set_delta(i);
	x.set_lambda(j);
	x.set_etype(string_to_edge_type(et));
	return is;
}

}


/*!
    \fn DDG::INFO_EDGE::check()
    \returns true is the object seems ok.
Currant implementation always return true.
 */
bool DDG::INFO_EDGE::check()
{
    /// @todo implement me
	return true;
}


/*!
    \fn DDG::INFO_EDGE::set_write_register_id(int regtypeid)
    In case of a dependence through register, sets the regiser type of the result: it is allowed in case of all dependences through registers, plus reuse and killer edges. Returns true if the kind of the dependance allows to set a register type.
 */
bool DDG::INFO_EDGE::set_register_type_id(int regtypeid)
{
    if ((this->is_dep_reg()) || (kind==killerdep) || (kind==reusedep))  {
    	regtype_id = regtypeid;
    	return true;
    }
    else return false;
    
}


/*!
    \fn DDG::INFO_EDGE::is_flow_reg()
    Returns true if the given edge is a flow dep through a register
 */
bool DDG::INFO_EDGE::is_flow_reg() const
{
    if(kind==flowdep_reg) return true;
    else return false;
}


/*!
    \fn DDG::INFO_EDGE::is_flow_reg(int regtypeid)
    Returns true if the given edge is a flow dep through a register of a given type
 */
bool DDG::INFO_EDGE::is_flow_reg(int regtypeid) const
{
    if(kind==flowdep_reg && regtype_id==regtypeid) return true;
    else return false;
}


/*!
    \fn DDG::INFO_EDGE::is_dep_reg()
    \returns true if the edge is a flowdep_reg, and antidep_reg, an outputdep_reg, inputdep_reg. False otherwise.
 */
bool DDG::INFO_EDGE::is_dep_reg() const
{
    if ((kind== flowdep_reg) || (kind==antidep_reg) || (kind==outputdep_reg) || (kind == inputdep_reg)) return true;
    else return false;
}
