/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/
#ifndef __info_edge_H
#define __info_edge_H
#include <iostream>
#include <string> 
/*! \file info_edge.h
\brief Header file of INFO_EDGE class.
Edges may contain many useful information, such as: latencies, distances, type of dependence, etc. 
*/
using namespace std;
namespace DDG{
class INFO_NODE;
class INFO_EDGE;
class DAG;
class LOOP;

/******************** List of Types *****************/

/*! \var typedef enum edge_type
 \brief Edges corresponds to data dependences. Data dependences are classified into multiple types : flowdep_reg, antidep_reg, outputdep_reg, inputdep_reg,  flowdep_mem, etc. 
 \remarks The user can modify this set of edge types in order to take into account user-defined data dependences.
 */
typedef enum  {
flowdep_reg=1, // flow dependence through register
antidep_reg, // WAR dependence through register
outputdep_reg, // WAW dependence through register
inputdep_reg, // RAR dependence through register
flowdep_mem, // flow dependence through memory
antidep_mem, // WAR dependence through memory
outputdep_mem, // WAW dependence through memory
inputdep_mem, // RAR dependence through memory
spilldep_mem, //Spill memory dependence arc
other_mem, // other dependence through memory
serial, // any serial arc
killerdep,  // Register consumer node to its killer node.
reusedep,  // Register reuse edge, from a killer to a producer node.
} edge_type; /* flow if flow dep through register */
/**  \brief Convert an edge_type to string ("flow", "serial", "antidep", ...).
 * \warning Returns empty string in case of unknownn edge_type
 */
std::string  edge_type_to_string(const edge_type);

/**  \brief Convert a string to an edge_type.
 * \pre The input string should be one of ("flow", "serial", "antidep", "outputdep, "inputdep") 
 */
edge_type  string_to_edge_type(const std::string);

/*! \class INFO_EDGE
\brief This class represents the attributes (information) attached to an edge in a DDG. 
\author Sid Touati		

Each edge \f$ e \f$ (data dependence or serial edge) has basically three attributes : 
 - Its type (flow, serial, antidep, ...)
 - Its integer latency : \f$ \delta(e) \f$
 - Its distance in terms of loop iterations (equal to 0 in case of a DAG) : : \f$ \lambda(e) \f$
*/
class INFO_EDGE
{
  friend class DDG::DAG;
  friend class DDG::LOOP;
  edge_type kind;
  int duration;
  int distance;
  int regtype_id;
  double val_duration; //!< \internal Used by critical cycle computation = (double) duration*/
  double sigma() const//!< \internal used by critical cycle computation */
  {return val_duration;}
  void set_sigma(double val)//!< \internal used by critical cycle computation */
  {val_duration=val;}
  
  public:
	/** \name Creation */
	//@{
	/** \brief Constructor by default. 
	 * 
	 * It sets the latency and the distance to 0 and the edge type to serial */ 
	INFO_EDGE() ;

	 /** \brief  Constructor with edge type. 
	  * 
	  * The latency and the distance are set to 0. */
	INFO_EDGE(edge_type et) ;

	/** \brief Constructor with latency, distance and edge type. */
	INFO_EDGE(int dur, int dis, edge_type et);

	/** \brief  Constructor with latency and edge type. 
	 * 
	 * The distance is set to 0.*/
	INFO_EDGE(int dur, edge_type et);

	
	INFO_EDGE(const INFO_EDGE &ie);

	//@}
	
	/** \name Access Operations */
	//@{
  	/** \brief Returns true if the edge is a flow dependence (throug registers or memory).*/
	bool is_flow() const
	{if (kind==flowdep_reg || kind==flowdep_mem) return true; else return false;}
	


	/**  \brief Returns the type of the edge (flow, serial, antidep, ...).*/
  	edge_type etype() const
  	{return kind;}
		
	/** \brief Returns the latency of the edge.*/
  	int delta() const
  	{return duration;}
	
	/** \brief Returns the distance of the edge in terms of iterations.*/
  	int lambda() const
  	{return distance;}
	/** \brief A simple check method that returns true if the object seems ok. */
	bool check();
	/** \brief Returns the register type id of the flowdep_reg edge. */
	int get_register_type_id() const {return regtype_id;}
	/** \brief Returns true if the edge is a flowdep_reg on the given register type. */
	bool is_flow_reg(int regtypeid) const;
	/** \brief Returns true if the edge is a flowdep_reg  */
	bool is_flow_reg() const;
	/** \brief Returns true if the edge is a dependence on a register (all type of dependences on registers).  */
	bool is_dep_reg() const;
	//@}
	
	/** \name Update Operations */
	//@{
	
	/** \brief Sets the edge type.*/
	void set_etype( const edge_type& et )
	{
		kind = et;
	}
	
	/** \brief  set a register type for the edge if it is a dependence through registers. */
	bool set_register_type_id(int regtypeid);

	/** \brief Sets the edge latency.*/
  	void set_delta(int val)
	{duration=val; val_duration=(double) val;}
	
	/** \brief Sets the edge distance.*/
  	void set_lambda(int lambda) 
	{distance=lambda;}
	
	//@}

	/** \name Input/Output stream operations  */
	//@{
	/** Output stream of the edge attribute. Prints \f$ \delta(e) \lambda(e) \f$ and its edge_type as a string ("flow", "serial, ...).
	 * 
	 */
	friend ostream& operator<<(ostream &os,  INFO_EDGE x) ; 
	
	/** Input stream of the edge attribute. Reads \f$ \delta(e) \lambda(e) \f$ and its edge_type as a string ("flow", "serial, ...).
	 * 
	 */
	friend istream& operator>>(istream &is, INFO_EDGE & x);  
    

	//@}
};


/*!\relates INFO_EDGE
	 * Returns True if two edge attributes are assumed equal : they should have the same latency (\f$ \delta \f$)) and the same distance (\f$ \lambda \f$).
 */
bool operator==(const INFO_EDGE, const INFO_EDGE);
}
#endif
