/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/
 
#ifndef __info_node_H
#define __info_node_H

#include "instructions_types.h"
#include <string> 
/*! \file info_node.h
\brief Header file of INFO_NODE class.
Nodes may contain many useful information, such as: unique identifiers, instruction type, latency, etc.
*/
using namespace std;
namespace DDG{
class INFO_NODE;
class INFO_EDGE;
class DAG;
class LOOP;

/*! \var typedef enum NodeFlag
 \brief Nodes may have some flag representing, special nodes. If a node corresponds to a regular instruction, then its flag is NodeFlag_Nothing.
 
 */

typedef enum {
  NodeFlag_Nothing = 0x0,	  // Operation has no flag
  NodeFlag_PartialDef = 0x1,       // Operation is partial definition.
  NodeFlag_SpillCode = 0x2,        // Operation is spill code.
  NodeFlag_Volatile = 0x4,         // Operation has volatile effects.
  NodeFlag_SPUpdate = 0x8,         // Operation is a SP update.
  NodeFlag_Prefetch = 0x10,        // Operation is a memory prefetch.
  NodeFlag_Preload = 0x20,         // Operation is a memory preload.
  NodeFlag_Barrier = 0x40,         // Operation is a memory barrier.
  NodeFlag_Clobber = 0x80,         // Operation has a clobber RegisterSet.
  NodeFlag_Hoisted = 0x100,        // Operation has been hoisted from loop.
  NodeFlag_DeadCode = 0x200,       // Operation is dead code.
  NodeFlag_SafeAccess = 0x800,     // Operation is safe for side-effects.
  NodeFlag_SafePerfs = 0x1000,     // Operation is safe for performances.
  NodeFlag_EntryCode = 0x4000,     // Operation is part of Procedure entry.
  NodeFlag_ExitCode = 0x8000,      // Operation is part of Procedure exit.
  NodeFlag_KillOp = 0x5000,       // Operation is a kill node
} NodeFlag;


/*! \class INFO_NODE
\brief This class represents the attributes (information) attached to each node in a DDG. 

\author Sid Touati		
		       
Each node (instruction) in a DDG has bsically two attributes : 
 - an C++ INSTRUCTIONS_TYPES object describing its generic instruction. 
 - an integer identifier that uniquely defines the node.
 * 
 * In order to avoid possible bugs, the user should not try to set the internal integer node identifier. Such unique identifier is automatically created internally by the %DDG library when creating a node.
*/

class INFO_NODE {
  friend class DDG::DAG;
  friend class DDG::LOOP;
  
  int identifier;
	std::string text; // the textual representation of the instruction, for instance "X=R+455". Not currently used
  INSTRUCTIONS_TYPES optype;
  NodeFlag node_flag;
    
  /**  \internal Assigns an integer as the identifier in the node attribute. This method is dangerous if the provided id is not unique.        */ 
  void set_id(int id)   {identifier=id;}
  
 public:
	 /** \name Creation  */
	//@{
	 /** \brief Default constructor.
	  * It constructs a node attribute with an instruction type "add_op" and and integer identifier equal to 0.  
	 */
	INFO_NODE() ;

	 
	 /** Constructor with two arguments 
	 * \param[in] id An integer identifier.
	 * \param[in] opt The INSTRUCTIONS_TYPES object describing the generic instruction of the node.
	 */
	INFO_NODE(int id, INSTRUCTIONS_TYPES opt);

	INFO_NODE(const INFO_NODE &in);

	/** Constructor with three arguments
	 * \param[in] id An integer identifier.
	 * \param[in] opt The INSTRUCTIONS_TYPES object describing the generic instruction of the node.
	 * \param[in] inst_text The textual representation of the instruction
	 */
	INFO_NODE(int id, INSTRUCTIONS_TYPES opt, std::string inst_text);
		   
	//@}
	 
	 /** \name Access Operations  */
	//@{
	/** \brief Returns OK if the object seems ok. */
	bool check();
	
	 /** \brief Returns the integer identifier of the node attribute. */
  	const int id() const
  	{return identifier;} 
	
	/** \brief Returns the instruction type of the node attribute. */
  	const  INSTRUCTIONS_TYPES instruction_type() const
  	{return optype;}

	/** \brief Returns the textual attribute of the node. */
	std::string get_string_attribute() const
	{return text;}
	/** \brief Returns  the id of tge instruction type of the node.
	 */
	int instruction_type_id() const
	{
		return optype.opcode_id();
	}	
	/** \brief Returns the flag of the node.*/	   
	NodeFlag get_node_flag() const
	{
		return node_flag;
	}
	//@}
		   
	/** \name Update Operations  */
	//@{
  	/** \brief Sets an INSTRUCTIONS_TYPES object in the node attribute. */
  	void set_instruction_type(INSTRUCTIONS_TYPES opt)
  	{optype=opt;}

	/** \brief Sets a textual attribue to the node.
	 * For instance, this textual attribute can be used to associate the textual code of the instruction.
	 * \remarks
	 * 	- Currently, a textual attribute is a single textual line. Hence, it should not contain the newline character \\n. If such character exists in the string, the current implementation of the method replaces it by a white space.
	 * 	- Double quotes characters " should be specialized. They should be written as \" inside the string attribute. */

	void set_string_attribute(std::string inst_text)
	{text=inst_text;}
	/** \brief  Set the flag of the node. */
	void set_node_flag ( const NodeFlag theValue )
	{
		node_flag = theValue;
	}
	//@}
	/** \name Input/Output stream operations  */
	//@{
	/** \brief Output stream of the node attributes in the form : node_id  INSTRUCTIONS_TYPES_attributes.*/
	 friend ostream& operator<<(ostream &os,  INFO_NODE x) ;
	 
	 /** \brief Input stream of the node attribute. Reads the node id and its INSTRUCTIONS_TYPES attribute.
	  */
	 friend istream& operator>>(istream &is, INFO_NODE & x);
   

    
	//@}
	
}
;  

}
#endif
