/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

/*! \file instructions_types.cpp
\brief Methods implementations of INSTRUCTIONS_TYPES class.
This file implements all methods accessing and modifying INSTRUCTIONS_TYPES objects. INSTRUCTIONS_TYPES objects correspond to the generic instructions of the ISA (ARCHITECTURE object).
*/
#include "instructions_types.h"


namespace DDG{

 
bool operator==(const INSTRUCTIONS_TYPES a, const INSTRUCTIONS_TYPES b) {
	if (a.opcode_id()==b.opcode_id()) return true;
	else return false;
}



/*!
    \fn INSTRUCTIONS_TYPES::nb_write_reg()
    \returns the number of register types written by the generic instuction.
 */
int INSTRUCTIONS_TYPES::nb_write_reg() const
{
    return list_reg_id.size();
}


/*!
    \fn INSTRUCTIONS_TYPES::get_list_regtypes()
    \returns the list of register types written by the generic instuction.
 */
leda::list<int> INSTRUCTIONS_TYPES::get_list_regtypes_id() const
{
    return list_reg_id;
}

ostream& operator<<(ostream &os,  INSTRUCTIONS_TYPES x)  /* output stream */
{
	os<<x.opcode()<<' '<<x.opcode_id()<<' '<<x.latency();
	return os;
}

istream& operator>>(istream &is, INSTRUCTIONS_TYPES & x) /* input stream */
{
	int opcode_id, lat;
	std::string opcode;
	is>>opcode>>opcode_id>>lat;
	x.set_opcode(opcode);
	x.set_opcode_id(opcode_id);
	x.set_latency(lat);
	return is;
}
INSTRUCTIONS_TYPES::INSTRUCTIONS_TYPES(const INSTRUCTIONS_TYPES &it)
{
	inst_type_id=it.inst_type_id;
	str_opcode= std::string(it.str_opcode);
	write_into_reg=it.write_into_reg;
	dr=it.dr;
	dw=it.dw;
	list_reg_id=it.list_reg_id;
	lat=it.lat;
}
	
INSTRUCTIONS_TYPES::INSTRUCTIONS_TYPES(std::string sopcode, int id):list_reg_id(), dw(0)
{
	inst_type_id=id;
	dr=0;
	str_opcode= std::string(sopcode);
	write_into_reg=false;
	lat=0;
}

INSTRUCTIONS_TYPES::INSTRUCTIONS_TYPES(const char *sopcode, int id, int latency) :list_reg_id(), dw(0)
{
	inst_type_id=id;
	dr=0;
	str_opcode= std::string(sopcode);
	write_into_reg=false;
	lat=latency;
}

INSTRUCTIONS_TYPES::INSTRUCTIONS_TYPES(std::string sopcode, int id, int latency) :list_reg_id(),  dw(0)
{
	inst_type_id=id;
	str_opcode= sopcode;
	dr=0;
	write_into_reg=false;
	lat=latency;
}


void DDG::INSTRUCTIONS_TYPES::set_delta_r(int dr_value)
{
	dr=dr_value;
	if (dr>=lat) {
	cerr<<"DDG Warning : the latency of the instruction type "<<str_opcode<<" is "<<lat<< " while its reading delay is "<<dr<<endl;
	}
}

bool DDG::INSTRUCTIONS_TYPES::set_delta_w(int regtype_id, int dw_value) 
{
	if(does_write_into_reg(regtype_id)){
		dw[regtype_id]=dw_value;
		return true;
	}
	else return false;
}

void DDG::INSTRUCTIONS_TYPES::set_latency(int latency_value) 
{
	lat=latency_value;
}


DDG::INSTRUCTIONS_TYPES::INSTRUCTIONS_TYPES(): list_reg_id(), dw(0)
{
	inst_type_id=0;
	str_opcode=std::string("");
	write_into_reg=false;
	dr=0;
	lat=1;
}


DDG::INSTRUCTIONS_TYPES::INSTRUCTIONS_TYPES(const char *sopcode, int id): dw(0),list_reg_id()
{
	inst_type_id=id;
	str_opcode= std::string(sopcode);
	write_into_reg=false;
	dr=0;
	lat=1;
}

}


/*!
    \fn DDG::INSTRUCTIONS_TYPES::add_write_reg_id(int)
    Adds a register type a writtent destination.
 */
void DDG::INSTRUCTIONS_TYPES::add_write_reg_id(int reg_id)
{
   write_into_reg=true;
   list_reg_id.push(reg_id);
}

void DDG::INSTRUCTIONS_TYPES::add_write_reg_id(int reg_id, int deltaw)
{
   write_into_reg=true;
   list_reg_id.push(reg_id);
   set_delta_w(reg_id, deltaw);
}


/*!
    \fn DDG::INSTRUCTIONS_TYPES::check()
    Simple check method.
    \returns true if the INSTRUCTIONS_TYPES object seems ok. False otherwise.
 */
bool DDG::INSTRUCTIONS_TYPES::check() const
{
    int regtype_id;
    if(write_into_reg){
    	forall(regtype_id, list_reg_id){
		if(dw[regtype_id] > lat){
		cerr << "DDG warning : instruction type check for the opcode "<<str_opcode<<" : the write latency is larger than the operation latency.\n" ;
		return false;
		}
    	}
    }
    if(dr > lat){
    	cerr << "DDG warning : instruction type check for the opcode "<<str_opcode<<" : the read latency is larger than the operation latency.\n" ;
    	return false;
    }
    	
    
    return true;
}


