/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

#ifndef __instructions_types_H
#define __instructions_types_H

/*! \file instructions_types.h
\brief A Documented file.
Details.
*/
#include <LEDA/core/list.h>
#include <LEDA/core/d_array.h>
#include "register_types.h"
using namespace std;
namespace DDG{


/*! \class INSTRUCTIONS_TYPES
\brief This class describes a generic instruction, used by the class ARCHITECTURE to describe an ISA. 
A generic instruction object contains opcode, unique id per opcode, string opcode, list of writtent registers, list of written register types, etc.

\author Sid Touati		
		     
* Each node in a data dependence graph corresponds to an instruction instance or to an operation. The generic instruction defined by the ISA is defined by the class INSTRUCTIONS_TYPES. A generic instruction is defined by a unique id, an opcode (of type string) and a flag saying if the instances of such generic instruction write into a register.
 * 
* To each node in a DAG or in a LOOP, we associate an attribute of type INSTRUCTIONS_TYPES describing its generic instruction description. A generic instruction has the following attributes:
 *  	- an integer latency
 * 	- an integer unique opcode id
 * 	- the string of the opcode (such as "add_op", "sub_op", etc.)
 * 	- a boolean flag saying wether the instruction writes into a register (as destination operand)
 * 	- an integer latency for reading delay from registers (\f$ \delta_r \f$)
 * 	- an integer latency for writing delay into registers (\f$ \delta_w \f$)

* The user can construct its own ISA as explained in \ref ISA_xml.
 */
class INSTRUCTIONS_TYPES{
	int inst_type_id;
	int lat; // hardware latency of the instruction in processor clock cycles
	int dr; // architecturally visible delay of reading from a register
	LEDA::d_array<int,int> dw; // architecturally visible delay of writing into a register 
	std::string str_opcode;
	bool write_into_reg;
	leda::list<int> list_reg_id;
	
	public:
		/**  \name Creation */
   		//@{
		/** \brief Constructor by default. 
		 * Build a generic instruction of type "add_op" by default, and gives it a null opcode id, and sets a flag saying that such instruction writes into a register. Reading and writing delays are set to zero. Latency set to 1.*/
		INSTRUCTIONS_TYPES() ;

		~INSTRUCTIONS_TYPES(){}
		
		/** Constructor with two arguments.
		* \param[in] sopcode The string of the opcode, such as "load_op", "store_op", etc.
		* \param[in] id A (unique) opcode id defining the instruction.
		 * Reading and writing delays are set to zero.
		*/
		INSTRUCTIONS_TYPES(std::string sopcode, int id);

		
		/** Constructor with two arguments.
		 * \param[in] sopcode The string (of type char *) of the opcode, such as "load_op", "store_op", etc.
		 * \param[in] id A (unique) opcode id defining the instruction.
		 * Reading and writing delays are set to zero.
		 */
		INSTRUCTIONS_TYPES(const char *sopcode, int id);

	
		/** Constructor with three arguments.
		 * \param[in] sopcode The string (of type char *) of the opcode, such as "load_op", "store_op", etc.
		 * \param[in] id A (unique) opcode id defining the instruction.
		 * \param[in] latency The latency of the instruction type (in processor clock cycles).
		 */
		INSTRUCTIONS_TYPES(const char *sopcode, int id,  int latency);


		/** Constructor with three arguments.
		 * \param[in] sopcode The string (of type std::string) of the opcode, such as "ld_op", "st_op", etc.
		 * \param[in] id A (unique) opcode id defining the instruction.
		 * \param[in] latency The latency of the instruction type (in processor clock cycles).
		 */
		INSTRUCTIONS_TYPES(std::string sopcode, int id, int latency);

		INSTRUCTIONS_TYPES(const INSTRUCTIONS_TYPES &it);

		//@}
		
		
		/**  \name Access Operations */
   		//@{
		/** \brief Returns the integer id that defines the instruction type.*/
		int opcode_id() const 
		{
			return inst_type_id;
		}

		/** \brief Returns \f$ \delta_r \f$ the architecturally visible latency of reading from a register (in processor clock cycles).*/
		int delta_r() const
		{
			return dr;
		}
		
    
		/** \brief Returns \f$ \delta_w \f$ the architecturally visible latency of writing into a register (in processor clock cycles).*/
		int delta_w(int regtype_id) const
		{
			if (dw.defined(regtype_id)&&is_value()) return dw[regtype_id];
			else return -1;
		}

		/** \brief Returns the latency of the instruction type (in processor clock cycles).*/
		int latency() const
		{
			return lat;
		}
	
	
		/** \brief Returns the string opcode of the instruction type */
		std::string opcode() const 
		{
			return str_opcode;
		}
		/** \brief Returns the string opcode of the instruction type */
		const char *opcode_str() const 
		{
			return (str_opcode.c_str());
		}
		/** \brief Returns true if instruction type writes into a register of any type. */
		bool is_value()  const
		{
			return write_into_reg;
		}
		/** \brief Returns the list of the id of the register types written by this generic instruction.*/
		leda::list<int> get_list_reg_id(){
			return list_reg_id;
		}
		/** \brief Same as get_list_reg_id() */
		
		leda::list<int> get_list_regtypes_id() const;
		
		/** \brief Returns the number of written register types. */
		int nb_write_reg() const;
		
		/** \brief Returns true if instruction type writes into a register of type t. */
		bool is_value(int t)  const
		{
			return does_write_into_reg(t);
		}
	
		/** \brief An alias of the method is_value(). Returns true if the instruction type writes into a register. */
		bool does_write_into_reg()  const
		{return is_value();}; 
		
		/** \brief Returns true if the instruction type writes into a register of a given type. */
		bool does_write_into_reg(int regtype_id)  const
		{
			return ( (list_reg_id.search(regtype_id)!=nil) && is_value());
		}; 
		/** \brief This method makes a simple check of the consistancy of the object. Returns true if OK, false otherwise. */
		bool check() const;
		//@}
			   
		/**  \name Update Operations */
   		//@{	   
		/** \brief Sets the opcode id */
		void set_opcode_id(const int& theValue) 
		{
		   inst_type_id = theValue;
		}

		/** \brief Sets the string opcode of the instruction type */
		void set_opcode(const std::string& theValue) 
		{
			str_opcode = theValue;
		}
	
		/** \brief Sets \f$ \delta_r \f$ the architecturally visible latency of reading from a register (in processor clock cycles).*/
		void set_delta_r(int dr_value);

	
		/** \brief Sets \f$ \delta_w \f$ the architecturally visible latency of writing into a register (in processor clock cycles).*/
		bool set_delta_w(int regtype_id, int dw_value) ;

		/** \brief Sets the latency of the instruction type (in processor clock cycles).*/
		void set_latency(int latency_value) ;
		/** \brief Add a written register type given by its id. */
		void add_write_reg_id(int t);
		/** \brief Add a written register type given by its id and sets \f$ \delta_w \f$. */
		void add_write_reg_id(int regtype_id, int deltaw);
		
		//@}
	/** \name Input/Output stream operations  */
	//@{
	/** Output stream of the INSTRUCTIONS_TYPES attribute. Print all INSTRUCTIONS_TYPES attributes in the form : string_opcode, opcode_id, latency, \f$ \delta_r \f$, \f$ \delta_w \f$,  does_write_into_reg?*/
	friend ostream& operator<<(ostream &os,  INSTRUCTIONS_TYPES x);
	
	 /** Input stream of the INSTRUCTIONS_TYPES. Read all INSTRUCTIONS_TYPES attributes in the following order : string_opcode, opcode_id, latency, \f$ \delta_r \f$, \f$ \delta_w \f$,  does_write_into_reg?*/
	friend istream& operator>>(istream &is, INSTRUCTIONS_TYPES & x);
	
	//@}
		
};


/*!\relates INSTRUCTIONS_TYPES
 * Returns True if two INSTRUCTIONS_TYPES objects are assumed equal : they are "equal" if they have the same opcode identifier.
 */

bool operator==(INSTRUCTIONS_TYPES a, INSTRUCTIONS_TYPES b);


}
#endif
