/***************************************************************************
 *   Copyright (C) 2009 by Touati Sid                                      *
 *   Sid-nospam.Touati-nospam@univ-cotedazur.fr                                      *
 *   Copyright  INRIA and the University of Versailles                     *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program (LGPL); if not, write to the          *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *   Exception : this software requires a licence  of the LEDA libary.     *
 *   If you are from academia, you can get a free leda binary licence 
 *   allowing you to use freely our DDG library.                           *
 *   Otherwise, you can buy a LEDA licence from algorithmic-solutions      *
 *   http://www.algorithmic-solutions.com                                  *
 *   LEDA binaries and source codes are excluded from this LGPL.           *
 *   Obtaining a LEDA licence is not mandatory to use our GPL software     *
 *   If the user wants, he can  create a free source-compatible replacement*
 *   for LEDA, and he is allowed to use our software under LGPL.           *
 ***************************************************************************/

/*! \file loop.h
\brief This file specifies the API of the LOOP class that describes cyclic data dependence graphs.

In optimizing compilation, simple internal loops are modeled by cyclic data dependence graphs. This library implements a C++ class called LOOP. The LOOP model we implement here is described in \ref loop_sec.
*/

/** \example loop_example.cpp
 * This is an example of how to use the LOOP class.
 * */


/*! \example gl2vcg.cpp
 A simple code example to show how to export a DDG file from gl format to vcg format to be visualized with xvcg
*/


#ifndef DDGLOOP_H
#define DDGLOOP_H

#include "DAG.h"

#include "LEDA/numbers/rational.h"
/**
\def Min(x,y)
\internal
Computes the minimum of \a x and \a y.
 */
#define Min(X,Y) X<Y ? X:Y


using namespace LEDA;
namespace DDG {
/*!  \class LOOP
	\brief This class represent cyclic data dependence graphs of simple loops (without branches). 
			       
       \author Sid Touati		
			       
	It inherits from the LEDA class GRAPH<INFO_NODE, INFO_EDGE>.  The attributes of the nodes are of type INFO_NODE, and the attributes of edges are of type INFO_EDGE. It models cyclic data dependence graphs. The implemented LOOP model is described in \ref loop_sec.
 * 
 * 	All LEDA access, creation and update methods can be used on a LOOP object.
 */
class LOOP: public DAG
{
	protected:
	
	/* internal functions for critical cycles */
	bool FLOYD(GRAPH<INFO_NODE, INFO_EDGE>&, node_matrix<double>&) const;
	double  MIN_CYCLE(GRAPH<INFO_NODE, INFO_EDGE>& ) const;
	double PREPARE_CRITICAL_CYCLE(GRAPH<INFO_NODE, INFO_EDGE>&, const LOOP&) const;
	double NEGATVE_CYCLE(GRAPH<INFO_NODE, INFO_EDGE>& , LEDA::list<edge>& negcycle_arcs) const;
	edge  theEdge(int source, int dest, int dist) const;  
	
	
	public:
	/** \name Creation */
	//@{
	 /** \brief Default constructor.
	  *
	  * Creates an empty LOOP object. A default ISA is assumed. */
	LOOP();
	  
	 void clear();
		 
	/** \brief Constructs an empty LOOP object encapsulating a user-defined generic description of the architecture (ISA).
	  * \param[in] isa ISA object, see ISA_xml.
	 * \throws InvalidISADesc When the ISA description file is corrupted, and object of class DDG::InvalidISADesc is thrown.
	*/
	LOOP(ARCHITECTURE isa);
	~LOOP();
	//@}
		
	/** \name Access Operations */
	//@{
	/** \brief Returns the maximal iteration edge distance in the loop. Formally equal to \f$ \max_{e \in E} \lambda(e)\f$.  */
	int MaxLambda() const;
	
	/** \brief Returns the minimal iteration edge distance in the loop. Formally equal to \f$ \min_{e \in E} \lambda(e)\f$.  */
	int MinLambda() const;
		
	/** \brief Returns the set of all \c flow edges (not necessarily through registers) from the source node \f$ u \f$ to the destination node \f$ v \f$. Formally equal to \f$ \{ e=(u,v)\in E | e \textrm {is a flow edge} \} \f$.
	   */
	LEDA::set<edge> flow_from_to(node u, node v) const;
	
	/** \brief Returns the set of all \c flow edges (not necessarily through registers) from the source node  to the destination node. Formally equal to \f$ \{ e=(u,v)\in E | e \textrm {is a flow edge} \} \f$.
	 * \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node. */
	LEDA::set<edge> flow_from_to(int id_u, int id_v) const;


	/** \brief Returns the set of all \c flow edges through registers from the source node \f$ u \f$ to the destination node \f$ v \f$. Formally equal to \f$ \{ e=(u,v)\in E_R \} \f$.
	 * \pre A unique register type exists.
	 * \throws NonUniqueRegisterType
	   */
	LEDA::set<edge> flow_reg_from_to(node u, node v) const;

	/** \brief Returns the set of all \c flow edges through registers  from the source node  to the destination node. Formally equal to \f$ \{ e=(u,v)\in E_R \} \f$.
	 * \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node. 
	 * \pre A unique register type exists.
	 * \throws NonUniqueRegisterType
	 */
	LEDA::set<edge> flow_reg_from_to(int id_u, int id_v) const;
	
	/** \brief Returns the set of all \c flow edges of type \f$ t \f$ from the source node \f$ u \f$ to the destination node \f$ v \f$. Formally equal to \f$ \{ e=(u,v)\in E_{R,t} \} \f$.
	 
	   */
	LEDA::set<edge> flow_reg_from_to(node u, node v, int t) const;

	/** \brief Returns the set of all \c flow edges through registers  from the source node  to the destination node. Formally equal to \f$ \{ e=(u,v)\in E_R \} \f$.
	 * \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node. 
	 * \param[in] t considered register type
	 */
	LEDA::set<edge> flow_reg_from_to(int id_u, int id_v, int t) const;
	

	/** \brief Returns the maximal distance between two nodes (in terms of number of iterations). Formally equal to \f$ \max_{e=(u,v)\in E} \lambda(e)\f$.
	 * \warning Returns -MAXINT if no edge exists between u and v.
	 */
	int max_dist(node u, node v) const;

	/** \brief Returns the maximal iteration distance between two nodes. Formally equal to \f$ \max_{e=(u,v)\in E} \lambda(e)\f$.
	 *  \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node.
	 * \warning Returns 0 if no edge exists between u and v.
	 * */
	int max_dist(int id_u, int id_v) const;
		
	/** \brief Returns the maximal iteration flow distance (not necessarily through registers) between two nodes. Formally equal to \f$ \max_{\textrm{a flow edge } e=(u,v)\in E} \lambda(e)\f$.
	 * \warning Returns -MAXINT if no edge exists between u and v.
	 */
	int max_flow_dist(node u, node v) const;
		
	/** \brief Returns the maximal iteration flow distance (not necessarily through registers) between two nodes. Formally equal to \f$ \max_{\textrm{a flow edge } e=(u,v)\in E} \lambda(e)\f$.
	 * \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node.
	 * \warning Returns -MAXINT if no edge exists between u and v.
	 * */
	int max_flow_dist(int id_u, int id_v) const;
		
	
	/** \brief Returns the maximal iteration flow distance through registers between two nodes. Formally equal to \f$ \max_{e=(u,v)\in E_R} \lambda(e)\f$.
	 * \warning Returns -MAXINT if no edge exists between u and v.
	 * \pre A unique register type exists in the ISA
	  \throw NonUniqueRegisterType
	 */
	int max_flow_reg_dist(node u, node v) const;
	
	/** \brief Returns the maximal iteration flow distance through registers  between two nodes. Formally equal to \f$ \max_{e=(u,v)\in E_R} \lambda(e)\f$.
	 * \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node.
	 * \warning Returns -MAXINT if no edge exists between u and v.
	 \pre A unique register type exists in the ISA
	 \throw NonUniqueRegisterType
	 * */
	int max_flow_reg_dist(int id_u, int id_v) const;
	
	/** \brief Returns the maximal iteration flow distance through registers of type t between two nodes. Formally equal to \f$ \max_{e=(u,v)\in E_{R,t}} \lambda(e)\f$.
	 * \warning Returns -MAXINT if no edge exists between u and v.

	 */
	int max_flow_reg_dist(node u, node v, int t) const;
	
	/** \brief Returns the maximal iteration flow distance through registers  between two nodes. Formally equal to \f$ \max_{e=(u,v)\in E_{R,t}} \lambda(e)\f$.
	 * \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node.
	 * \param[in] t considered register type
	 * \warning Returns -MAXINT if no edge exists between u and v.
	
	 * */
	int max_flow_reg_dist(int id_u, int id_v, int t) const;	

	/** \brief Returns the minimal iteration distance between two nodes. Formally equal to \f$ \min_{e=(u,v)\in E} \lambda(e)\f$.
	 * \warning Returns +MAXINT if no edge exists between u and v.
	 */
	int min_dist(node u, node v) const;
	
	/** \brief Returns the minimal iteration distance between two nodes. Formally equal to \f$ \max_{e=(u,v)\in E} \lambda(e)\f$.
	 *  \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node.
	 * \warning Returns +MAXINT if no edge exists between u and v.
	 * */
	int min_dist(int id_u, int id_v) const;
	
	/** \brief Returns the  minimal iteration flow distance  (not necessarily through registers)  between two nodes. Formally equal to \f$ \min_{{\textrm a flow edge }e=(u,v)\in E} \lambda(e)\f$.
	 * \warning Returns +MAXINT if no edge exists between u and v.
	
	 */
	int min_flow_dist(node u, node v) const;
	
	/** \brief Returns the  minimal iteration flow distance (not necessarily through registers)  between two nodes. Formally equal to \f$ \max_{\textrm{a flow edge }  e=(u,v)\in E} \lambda(e)\f$.
	 * \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node.
	 * \warning Returns +MAXINT if no edge exists between u and v.
	 * */
	int min_flow_dist(int id_u, int id_v) const;

	/** \brief Returns the minimial iteration flow distance through registers between two nodes. Formally equal to \f$ \min_{e=(u,v)\in E_R} \lambda(e)\f$.
	 * \warning Returns +MAXINT if no edge exists between u and v.
	 */
	int min_flow_reg_dist(node u, node v) const;
	
	/** \brief Returns the minimal iteration flow distance through registers  between two nodes. Formally equal to \f$ \min_{e=(u,v)\in E_R} \lambda(e)\f$.
	 * \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node.
	 * \warning Returns +MAXINT if no edge exists between u and v.
	 * */
	int min_flow_reg_dist(int id_u, int id_v) const;
	
	/** \brief Returns the minimial iteration flow distance through registers of type t between two nodes. Formally equal to \f$ \min_{e=(u,v)\in E_{R,t}} \lambda(e)\f$.
	 * \warning Returns +MAXINT if no edge exists between u and v.
	 */
	int min_flow_reg_dist(node u, node v, int t) const;
	
	/** \brief Returns the minimal iteration flow distance through registers of type t between two nodes. Formally equal to \f$ \min_{e=(u,v)\in E_{R,t}} \lambda(e)\f$.
	 * \param[in] id_u The integer identifier of the source node.
	 * \param[in] id_v The integer identifier of the destination node.
	 * \param[in] t considered register type
	 * \warning Returns +MAXINT if no edge exists between u and v.
	 * */
	int min_flow_reg_dist(int id_u, int id_v, int t) const;
			   
	/** \brief Returns the ratio of the critical cycle of the loop.
	 *
	 * The critical cycle of a loop is defined by the one maximizing the ratio \f$ \max_{C\textrm{ a cycle}} \frac{\sum_{e \in C} \delta(e)} {\sum_{e \in C} \lambda(e)}\f$. Contrary to what it is usually implemented, this method can detect null cycles, \c i.e., cycles \f$ C \f$ such as \f$ \sum_{e \in C} \delta(e) =0 \f$ and \f$ \sum_{e \in C} \lambda(e) \ge 0 \f$. This method returns -1 if no cycle exists in the graph.
	 * \pre All cycles should have nonnegative distance, that is \f$ \forall C \textrm{ a cycle in } G, \sum_{e \in C} \lambda(e)\ge 0 \f$. Note that the edges may have nonpositive latencies (\f$ \delta(e) \f$ may be positive or negative).
	* \remarks
	 * 	- The implemented algorithm is the one published by E.L Lawler in Periodic Optimization (1972, Springer Verlag). Its complexity is \f$ O(n\times m \times \log(n)) \f$, where \f$ n \f$ is the number of nodes and \f$ m \f$ is the number of edges.
	 * 	- Since the original algorithm does not try to detect null cycles, we added our own implementation mechanism to do it.
	 * */
		
	LEDA::rational CRITICAL_CYCLE() const;
		
	/** Returns the ratio of the critical cycle of the loop and gives its list of edges.
	 *
	 * A critical cycle in a loop data dependence graph is defined as the one which maximizes the ratio \f$ \max_{C\textrm{ a cycle}} \frac{\sum_{e \in C} \delta(e)} {\sum_{e \in C} \lambda(e)} \f$. Contrary to what it is usually implemented, this method can detect null cycles, \c i.e., cycles \f$ C \f$ such as \f$ \sum_{e \in C} \delta(e) =0 \f$ and \f$ \sum_{e \in C} \lambda(e) \ge 0 \f$. This method returns -1 if no cycle exists in the graph.
	\param[out] el Contains the list of edges that belongs to the critical cycle.
	 * \warning If the critical cycle has a null latency  (\f$ \sum_{e \in C} \delta(e) =0 \f$)  and a nonnegative distance \f$ \sum_{e \in C} \lambda(e) > 0 \f$, then the output list of edges nl is empty and the method returns zero. However, our method is able to build the list of edges belonging to a critical cycle if  \f$ \sum_{e \in C} \delta(e) =  \sum_{e \in C} \lambda(e) =0\f$.
	 * \pre All cycles should have nonnegative distance, that is \f$ \forall C \textrm{ a cycle in } G, \sum_{e \in C} \lambda(e)\ge 0 \f$. Note that the edges may have nonpositive latencies (\f$ \delta(e) \f$ may be positive or negative).
	 *
	 \remarks
		- The implemented algorithm is the one published by E.L Lawler in Periodic Optimization (1972, Springer Verlag). Its complexity is \f$ O(n\times m \times \log(n)) \f$, where \f$ n \f$ is the number of nodes and \f$ m \f$ is the number of edges.
		- Since the original algorithm does not try to detect null cycles, we added our own implementation mechanism to do it.
		  */
	LEDA::rational CRITICAL_CYCLE(LEDA::list<edge> &el) const;
	
	    
	/**  \brief Returns the distance \f$ \lambda(e) \f$ of the edge \f$ e \f$ (in terms of number of iterations).  */
	int lambda(edge e) const;	
					  
	//@}
			   
	/** \name Update Operations */
	//@{
	
	/**  \brief Duplicates a LOOP.
	* \param[in] G2 The current LOOP object will contain a duplicated copy of G2.
	*/
	void copy(const LOOP &G2);
		
	/** \brief Duplicates a LOOP.
		\param[in] G2 The current LOOP object will contain a duplicated copy of  G2.
	    * \param[out] mn mn is a mapping from the set of nodes of G2 to the set of nodes of the current LOOP. \f$ \forall u\in G_2, mn[u] \f$ contains the node belonging to the current LOOP corresponding to its original copy \f$ u \in G_2 \f$.
	    * \param[out] me me is a mapping from the set of edges of G2 to the set of edges of the current LOOP. \f$ \forall e \in G_2, me[e] \f$ contains the edge belonging to the current LOOP corresponding to its original copy \f$ e \in G_2 \f$.
	\remarks
		   LOOP::build_internal_structures() is called inside this method.
	 */
	   void copy(const LOOP& G2, node_map<node> &mn, edge_map<edge> &me);
		
	
		
			   
	/** \brief Build internal data structures for the LOOP object. 
	   *
	   * For internal information consistency, this function member should be called after update operations such as LOOP::new_node(INSTRUCTIONS_TYPES it), LOOP::new_edge(INFO_EDGE), etc. This method compute some internal information returned by other methods (the set \f$ V_R \f$, the set of consumers per value, etc.). If the LOOP object is not modified after calling build_internal_structures(), all its internal information remain consistent.
		remarks For efficiency, the user can group series of update operations before ultimately calling build_internal_structures() 
	 */
	void build_internal_structures();
		
		

	
	


		
	/** \brief Sets a distance \f$ \lambda(e) \ge 0 \f$ (in terms of number of iterations) to the edge \f$ e \f$ .  */
	void set_lambda(edge e, int l);

		 
	

	//@}
	   /** Import and Export Methods */
	  //@{
		
	   /** \brief Exports the LOOP DDG to a vcg file in order to be visualized with the xvcg tool. 
	* For more information about xvcg, see http://rw4.cs.uni-sb.de/users/sander/html/gsvcg1.html
	    * The user can navigate using vcg on all DDG attrbutes (types of edges, opcodes, etc.)
	 * \param[in] filename the file name of the DDG in vcg format
	   */
	void write_to_vcg(const char *filename) const;

	/** \brief Exports the LOOP DDG to a vcg file in order to be visualized with the xvcg tool. This function considers one register type.
	 * \param[in] filename the file name of the DDG in vcg format
	 * \param[in] t  the considered register type
	 * For more information about xvcg, see http://rw4.cs.uni-sb.de/users/sander/html/gsvcg1.html
	 * The user can navigate using vcg on all DDG attrbutes (types of edges, opcodes, etc.)
	 */
	void write_to_vcg(const char *filename, int t) const;

	/** \brief Exports the LOOP to a gl file. If multiple registers types exist, they are all considered as the same type.
	 * \param[in] filename the file name of the DDG in gl format
	*/
	void write_to_gl(const char *filename) const;
	
	/** \brief Exports the LOOP to a gl file considering one register type.
	 * \param[in] filename the file name of the DDG in gl format
	 * \param[in] regtype_id All flow_reg edges through this register type are output to gl as flow_reg. Other flow_reg edges are output as "serial" edges.*/
	
	void write_to_gl(const char *filename, int regtype_id) const;
		
	/** \brief Imports a DDG loop from a gl file. In case of multiple registers types, this function allows to considere one of them.
	 * \param[in] filename the file name of the DDG in gl format
	 * \param[in] regtype_id The flow_reg edges in the gl file are considered as flow_reg through this register type.
	 \returns true if OK
	\remarks
		 
		   - LOOP::build_internal_structures() is called inside this method.
	    * \throws BadIODDG If an IO error occurs (corrupted or non present input files for instance), an exception object of type DDG::BadIODDG is thrown
	 *  * \pre Before reading a LOOP from a gl file, it is assumed that the ISA of the LOOP defines all the opcodes ids of the nodes as present in the gl file.  The constructor DDG::LOOP(ARCHITECTURE) defines a LOOP with a user defined ISA, otherwise a default ISA is assumed, see  \ref ISA_xml.
	 * */
	bool read_from_gl(const char *filename, int regtype_id);
	
	
	/** \brief Imports a DDG loop from a gl file.
	* \param[in] filename the file name of the DDG in gl format
	\returns true if OK.
	\pre A unique register type should exist in the architecture.
	\pre Before reading a LOOP from a gl file, it is assumed that the ISA of the LOOP defines all the opcodes ids of the nodes as present in the gl file.  The constructor DDG::LOOP(ARCHITECTURE) defines a LOOP with a user defined ISA, otherwise a default ISA is assumed, see  ISA_xml.
	\remarks
		 
		   - LOOP::build_internal_structures() is called inside this method.
	    * \throws BadIODDG If an IO error occurs (corrupted or non present input files for instance), an exception object of type DDG::BadIODDG is thrown
	 * */
	bool read_from_gl(const char *filename);

	/** \brief Exports the LOOP to a leda XML format. .*/
	
	void write_to_xml(const char *filename) const;
	
	/** \brief Imports  the LOOP from an XML file.
	\returns true if ok.
  	 */
	bool read_from_xml(const char *filename);
    	bool check();
	bool check(int regtype_id);
	
	//@}
	
};

/* Some nice loop manipulation functions. */

/*! \fn void unroll(const LOOP& L, DAG &unrolled, int unroll_degree);
\brief Returns the body of the unrolled loop.
\param[in] L The loop to be considered. Not altered by this function.
\param[out] unrolled The resulting DAG (the loop body of the unrolled loop).
\param[in] unroll_degree The degree of loop unrolling. If equal to zero, no loop unrolling is performed and the resulting DAG is empty. 
\remarks 
	- The degree unrolling tells how much loop bodies are duplicated. If the degree of unrolling is equal to 1, the resulting DAG contains only one loop body.
	- This unrolling function preserves and handles recurrent edges.
*/


void unroll(const LOOP& L, DAG &unrolled, int unroll_degree);

/*! \fn void unroll(const LOOP& L, LOOP &unrolled, int unroll_degree);
	\brief Returns an unrolled loop.
	\param[in] L The loop to be considered. Not altered by this function.
	\param[out] unrolled The resulting unrolled loop. 
	\param[in] unroll_degree The degree of loop unrolling. If equal to zero, no loop unrolling is performed and the resulting loop is empty. 
	\remarks
	- The degree unrolling tells how much loops bodies are duplicated. If the degree of unrolling is equal to 1, the resulting loop contains only one copy of the initial loop.
	- This unrolling function preserves and handles recurrent edges.
 * \pre Unrolling degree and all dependence distances should be nonnegative. \f$ \forall u\in V, \lambda(u)\ge 0 \f$.
*/

void unroll(const LOOP& L, LOOP &unrolled, int unroll_degree);


/*! \fn void loop_merge(const LOOP &L1, const LOOP &L2, LOOP &L3);
\brief Merges two independent loop DDGs in order to produce a new DDG.
\param[in] L1 The first DDG to consider.
\param[in] L2 The second DDG to consider.
\param[out] L3 A new loop consisting of merging distinct copies of L1 and L2.
 * \pre Unrolling degree and all dependence distances should be nonnegative. \f$ \forall u\in V, \lambda(u)\ge 0 \f$.
*/
void loop_merge(const LOOP &L1, const LOOP &L2, LOOP &L3);

/*! \fn bool retime_ddg(LOOP &G, LEDA::node_array<int> &r);
	\brief Computes and applies a valid loop retiming (called also loop shifing).
	\param[in,out] G The loop to retime (modified by the loop retiming)
	\param[out] r the computed retiming.
	\returns true if a valid retiming has been found.
 * 	\returns false if no valid retiming has been found (in case of the presence of nonpositive distance cycle).
 *
 * Loop retiming consists of the following graph transformation: for each loop statement \f$ u \f$, we associate an integer shift \f$ r(u) \f$ which means that we delay the operation \f$ u \f$ by \f$ r(u) \f$ iterations. Then, each statement \f$ u \f$ that was representing the operation \f$ u \f$  of loop iteration \f$ i \f$ represents now the operation \f$ u \f$  of loop iteration \f$ i-r(u) \f$. The new distance of each arc \f$ e=(u,v) \f$ becomes equal to  \f$ \lambda_r(e)=\lambda(e)+r(v)-r(u) \f$. Such new distance is always nonnegative \f$ \lambda_r(u)\ge 0, \forall u \in V \f$.

 For more details about loop retiming, please study the excellent research article of Leiserson and Saxe published in Algorithmica 1991 (Retiming asynchronous circuits).
 * \warning If the retiming is not valid, the loop would been modified anywhere.
 * 
*/
bool retime_ddg(LOOP &G, LEDA::node_array<int> &r);

/*! \fn bool retime_ddg(LOOP &G);
	\brief Computes and applies a valid loop retiming (called also loop shifing).
	\param[in,out] G The loop to retime (modified by the loop retiming)
	\returns true if a valid retiming has been found.
 * 	\returns false if no valid retiming has been found (in case of the presence of nonpositive distance cycle).
 *
 * Loop retiming consists of the following graph transformation: for each loop statement \f$ u \f$, we associate an integer shift \f$ r(u) \f$ which means that we delay the operation \f$ u \f$ by \f$ r(u) \f$ iterations. Then, each statement \f$ u \f$ that was representing the operation \f$ u \f$  of loop iteration \f$ i \f$ represents now the operation \f$ u \f$  of loop iteration \f$ i-r(u) \f$. The new distance of each arc \f$ e=(u,v) \f$ becomes equal to  \f$ \lambda_r(e)=\lambda(e)+r(v)-r(u) \f$. Such new distance is always nonnegative \f$ \lambda_r(u)\ge 0, \forall u \in V \f$.

 For more details about loop retiming, please study the excellent research article of Leiserson and Saxe published in Algorithmica 1991 (Retiming asynchronous circuits).
 * \warning If the retiming is not valid, the loop would been modified anywhere.
 *
*/

bool retime_ddg(LOOP &G);

/*! \fn bool apply_retiming(LOOP &G, const LEDA::node_array<int> r);
	\brief Applies the loop retiming given as input.
	\param[in,out] G The loop to retime (modified by the loop retiming)
 * 	\param[in] r the input retiming given by the user.
 * 	
	\returns true if the input retiming is valid (all retimed edge distances are nonnegative).
 *\returns false if the input retiming is not valid (in case of the presence of a nonpositive retimed edge distance).
 * \warning If the retiming is not valid, the loop would been modified anywhere.
 *
 * Loop retiming consists of the following graph transformation: for each loop statement \f$ u \f$, we associate an integer shift \f$ r(u) \f$ which means that we delay the operation \f$ u \f$ by \f$ r(u) \f$ iterations. Then, each statement \f$ u \f$ that was representing the operation \f$ u \f$  of loop iteration \f$ i \f$ represents now the operation \f$ u \f$  of loop iteration \f$ i-r(u) \f$. The new distance of each arc \f$ e=(u,v) \f$ becomes equal to  \f$ \lambda_r(e)=\lambda(e)+r(v)-r(u) \f$. Such new distance is always nonnegative \f$ \lambda_r(u)\ge 0, \forall u \in V \f$.

 For more details about loop retiming, please study the excellent research article of Leiserson and Saxe published in Algorithmica 1991 (Retiming asynchronous circuits).
 *
*/
bool apply_retiming(LOOP &G, const LEDA::node_array<int> r);



/*! \fn bool is_lexicographic_positive(const LOOP G&, edge e);
	\brief Checks if all circuits of \f$ G \f$ are lexicographic positive. That is, \f$ \forall C \f$ a circuit in \f$ G \f$, \f$ \lambda(C)=\sum_{e\in C} \lambda(e) > 0 \f$.
	\param[in] G The loop check
 * 	\param[out] e An edge belonging to a non positive circuit, in case when the DDG is not lexicographic positive.
 * 	
	\returns true if the input DDG is lexicographic positive. False otherwise. 
	
 *
 * In order to check if a DDG is lexicographic positive, we use the lemma number 10.3 in page 193 of the PhD thesis of M. Sid Touati. That lemma uses loop retiming, but the current implementation converts loop retiming to a potential computation algorithms. As proved by Leiserson and Saxe, every retiming is equivalent to a potential in a graph.
 *
*/

bool is_lexicographic_positive( LOOP &G,  edge &e);

/*! \fn is_lexicographic_positive( LOOP &G,  leda::list<edge>  &le);
	\brief Checks if all circuits of \f$ G \f$ are lexicographic positive. That is, \f$ \forall C \f$ a circuit in \f$ G \f$, \f$ \lambda(C)=\sum_{e\in C} \lambda(e) > 0 \f$.
	\param[in] G The loop check
 * 	\param[out] le A list of a delinquent circuit \f$ C \f$, that  is \f$ \lambda(C) \le 0 \f$.
 * 	
	\returns true if the input DDG is lexicographic positive. False otherwise. 
	
 *
 * In order to check if a DDG is lexicographic positive, we use the lemma number 10.3 in page 193 of the PhD thesis of M. Sid Touati. That lemma uses loop retiming, but the current implementation converts loop retiming to a potential computation algorithms. As proved by Leiserson and Saxe, every retiming is equivalent to a potential in a graph.
 *
 */
bool is_lexicographic_positive( LOOP &G,  leda::list<edge>  &le);

}


#endif
